# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, 2021 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'Haiku BFS Attributes'
PLUGIN_AUTHOR = 'Philipp Wolfer'
PLUGIN_DESCRIPTION = 'Save and load metadata to/from Haiku BFS attributes.'
PLUGIN_VERSION = "1.2"
PLUGIN_API_VERSIONS = ["2.2", "2.3", "2.4", "2.5", "2.6"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"


import os
import struct
import sys
from collections import namedtuple
from ctypes import (CDLL, POINTER, Structure, byref, c_char_p, c_int, c_long,
                    c_longlong, c_size_t, c_ssize_t, c_uint32, c_void_p, cast)
from ctypes.util import find_library
from functools import partial

from picard import log
from picard.file import (
    register_file_post_load_processor,
    register_file_post_save_processor,
    )
from picard.util import thread

if sys.platform == 'haiku1':

    class AttrInfo(Structure):
        _fields_ = [
            ('type', c_uint32),
            ('size', c_size_t)]

        def __repr__(self):
            return 'AttrInfo(type=%r, size=%r)' % (self.type, self.size)

    try:
        libbe_path = find_library('be') or 'libbe.so'
        be = CDLL(libbe_path, use_errno=True)
        be.fs_stat_attr.restype = c_int
        be.fs_stat_attr.argtypes = [c_int, c_char_p, POINTER(AttrInfo)]
        be.fs_read_attr.restype = c_ssize_t
        be.fs_read_attr.argtypes = [c_int, c_char_p, c_uint32, c_size_t,
                                    c_void_p, c_size_t]
        be.fs_remove_attr.restype = c_int
        be.fs_remove_attr.argtypes = [c_int, c_char_p]
        be.fs_write_attr.restype = c_ssize_t
        be.fs_write_attr.argtypes = [c_int, c_char_p, c_uint32, c_size_t,
                                     c_void_p, c_size_t]
    except OSError:
        log.error('haikuattrs: unable to load libbe', exc_info=True)
        be = None
else:
    log.warning('haikuattrs: this plugin is only for the Haiku operating system')
    be = None

if be:
    Attr = namedtuple('Attr', ['type', 'name'])
    attr_map = {
        'artist'     : Attr(b'CSTR', b'Audio:Artist'),
        'album'      : Attr(b'CSTR', b'Audio:Album'),
        'title'      : Attr(b'CSTR', b'Media:Title'),
        'date'       : Attr(b'LONG', b'Media:Year'),
        'comment:'   : Attr(b'CSTR', b'Media:Comment'),
        'tracknumber': Attr(b'LONG', b'Audio:Track'),
        'genre'      : Attr(b'CSTR', b'Media:Genre'),
        'composer'   : Attr(b'CSTR', b'Audio:Composer'),
        '~rating'    : Attr(b'LONG', b'Media:Rating'),
        '~length'    : Attr(b'LLNG', b'Media:Length'),
        '~bitrate'   : Attr(b'CSTR', b'Audio:Bitrate'),
    }

    def get_numeric_type(attr):
        return struct.unpack('>I', attr.type)[0]

    def read_attr(fd, attr):
        info = AttrInfo()
        if be.fs_stat_attr(fd, attr.name, byref(info)) == -1:
            return None

        buffer = b'\0' * info.size
        bytes_read = be.fs_read_attr(fd, attr.name, info.type, 0,
                                     buffer, len(buffer))

        result = None
        if bytes_read > -1:
            buffer = buffer[:bytes_read]
            if attr.type == b'LONG':
                result = str(struct.unpack('=l', buffer)[0])
            elif attr.type == b'LLNG':
                result = str(struct.unpack('=q', buffer)[0])
            elif attr.type == b'CSTR':
                result = buffer.decode('utf-8', errors='replace')
            else:
                raise ValueError('Unsupported attribute type %s' % attr.type)

        log.debug("haikuattrs: fs_read_attr(%r, %r, %r, %r, %r, %r) -> %r: %r" % (
            fd, attr.name, info.type, 0, buffer, len(buffer), bytes_read, result))
        return result

    def remove_attr(fd, attr):
        return be.fs_remove_attr(fd, attr.name) == 0

    def write_attr(fd, attr, attr_value):
        if attr.type in (b'LONG', b'LLNG'):
            try:
                attr_value = int(attr_value)
            except ValueError:
                return False
            length = 4
            c_type = c_long if attr.type == b'LONG' else c_longlong
            buffer = byref(c_type(attr_value))
        elif attr.type == b'CSTR':
            attr_value = attr_value.encode('utf-8')
            length = len(attr_value)
            buffer = cast(attr_value, c_char_p)
        else:
            raise ValueError('Unsupported attribute type %s' % attr.type)

        int_type = get_numeric_type(attr)
        ret_val = be.fs_write_attr(fd, attr.name, int_type, 0, buffer, length)
        log.debug("haikuattrs: fs_write_attr(%r, %r, %r, %r, %r, %r) -> %r" % (
            fd, attr.name, int_type, 0, buffer, length, ret_val))
        return ret_val >= 0

    def get_attribute_value(tag, file):
        metadata = file.orig_metadata
        value = metadata[tag]
        if value:
            if tag == 'date':
                value = value[:4]
            elif tag == '~length':
                # Haiku expects the Media:Length attribute in nanoseconds
                value = metadata.length * 1000
            elif tag == '~bitrate':
                value = '%s kbps' % value
            elif tag == '~rating':
                # Haiku's ratings are internally repesented by integers between 1 and 10
                value = int(value * 2)
        return value

    def set_attrs_from_metadata(file):
        log.debug('haikuattrs: setting attributes for %s' % file.filename)
        fd = os.open(file.filename, os.O_RDWR)
        try:
            for tag, attr in attr_map.items():
                try:
                    value = get_attribute_value(tag, file)
                except (TypeError, ValueError) as err:
                    log.warning(
                        'haikuattrs: failed converting tag %s with value %r for %s: %r',
                        tag, value, file.filename, err)
                    continue
                if value:
                    if not write_attr(fd, attr, value):
                        log.error('haikuattrs: setting %s=%s for %s failed' % (
                            attr.name, value, file.filename))
                else:
                    remove_attr(fd, attr)
        finally:
            os.close(fd)

    def set_attrs_from_metadata_finished(file, result=None, error=None):
        if error:
            log.error('haikuattrs: setting attributes for %s failed: %r' % (
                file.filename, error))
        else:
            log.debug('haikuattrs: attributes set for %s' % file.filename)

    def load_attrs_to_metadata(file):
        fd = os.open(file.filename, os.O_RDONLY)
        filename, _ = os.path.splitext(file.base_filename)
        try:
            for tag, attr in attr_map.items():
                # Technical variables that are loaded directly from the file
                if tag in ('~bitrate', '~length'):
                    continue
                # Ignore tags for which metadata is included in the file.
                # But ignore the special case where the title has been set to
                # the filename.
                value = file.metadata[tag]
                if value and not (tag == 'title' and value == filename):
                    continue
                value = read_attr(fd, attr)
                if value:
                    file.metadata[tag] = value
                    file.orig_metadata[tag] = value
        finally:
            os.close(fd)

    def load_attrs_to_metadata_finished(file, result=None, error=None):
        if error:
            log.error('haikuattrs: loading attributes for %s failed: %r' % (
                file.filename, error))
        else:
            log.debug('haikuattrs: attributes loaded for %s' % file.filename)
            file.update()

    def on_file_load_processor(file):
        thread.run_task(
            partial(load_attrs_to_metadata, file),
            partial(load_attrs_to_metadata_finished, file))

    def on_file_save_processor(file):
        thread.run_task(
            partial(set_attrs_from_metadata, file),
            partial(set_attrs_from_metadata_finished, file))

    register_file_post_load_processor(on_file_load_processor)
    register_file_post_save_processor(on_file_save_processor)



import doctest


def load_tests(loader, tests, ignore):
    from plugins.addrelease import addrelease
    tests.addTests(doctest.DocTestSuite(addrelease))

    from plugins import decade
    tests.addTests(doctest.DocTestSuite(decade))

    from plugins.smart_title_case import smart_title_case
    tests.addTests(doctest.DocTestSuite(smart_title_case))

    from plugins.standardise_feat import standardise_feat
    tests.addTests(doctest.DocTestSuite(standardise_feat))

    from plugins.key_wheel_converter import key_wheel_converter
    tests.addTests(doctest.DocTestSuite(key_wheel_converter))

    from plugins import enhanced_titles
    tests.addTests(doctest.DocTestSuite(enhanced_titles))
    return tests



#!/usr/bin/env python3

from __future__ import print_function
import argparse
import os
import json
import zipfile

from hashlib import md5
from subprocess import check_call

from get_plugin_data import get_plugin_data

VERSION_TO_BRANCH = {
    None: '2.0',
    '1.0': '1.0',
    '2.0': '2.0',
}


def build_json(dest_dir):
    """
    Traverse the plugins directory to generate json data.
    """

    plugins = {}

    # All top level directories in plugin_dir are plugins
    for dirname in next(os.walk(plugin_dir))[1]:

        files = {}
        data = {}

        if dirname in [".git"]:
            continue

        dirpath = os.path.join(plugin_dir, dirname)
        for root, dirs, filenames in os.walk(dirpath):
            for filename in filenames:
                ext = os.path.splitext(filename)[1]

                if ext not in [".pyc"]:
                    file_path = os.path.join(root, filename)
                    with open(file_path, "rb") as md5file:
                        md5Hash = md5(md5file.read()).hexdigest()
                    files[file_path.split(os.path.join(dirpath, ''))[1]] = md5Hash

                    if ext in ['.py'] and not data:
                        data = get_plugin_data(os.path.join(plugin_dir, dirname, filename))

        if files and data:
            print("Added: " + dirname)
            data['files'] = files
            plugins[dirname] = data
    out_path = os.path.join(dest_dir, plugin_file)
    with open(out_path, "w") as out_file:
        json.dump({"plugins": plugins}, out_file, sort_keys=True, indent=2)


def zip_files(dest_dir):
    """
    Zip up plugin folders
    """

    for dirname in next(os.walk(plugin_dir))[1]:
        archive_path = os.path.join(dest_dir, dirname)
        archive = zipfile.ZipFile(archive_path + ".zip", "w")

        dirpath = os.path.join(plugin_dir, dirname)
        plugin_files = []

        for root, dirs, filenames in os.walk(dirpath):
            for filename in filenames:
                file_path = os.path.join(root, filename)
                plugin_files.append(file_path)

        if (len(plugin_files) == 1
            and os.path.basename(plugin_files[0]) != '__init__.py'):
            # There's only one file, put it directly into the zipfile
            archive.write(plugin_files[0],
                          os.path.basename(plugin_files[0]),
                          compress_type=zipfile.ZIP_DEFLATED)
        else:
            for filename in plugin_files:
                # Preserve the folder structure relative to plugin_dir
                # in the zip file
                name_in_zip = os.path.join(os.path.relpath(filename,
                                                           plugin_dir))
                archive.write(filename,
                              name_in_zip,
                              compress_type=zipfile.ZIP_DEFLATED)

        print("Created: " + dirname + ".zip")


# The file that contains json data
plugin_file = "plugins.json"

# The directory which contains plugin files
plugin_dir = "plugins"

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Generate plugin files for Picard website.')
    parser.add_argument('version', nargs='?', help="Build output files for the specified version")
    parser.add_argument('--build_dir', default="build", help="Path for the build output. DEFAULT = %(default)s")
    parser.add_argument('--pull', action='store_true', dest='pull', help="Pulls the remote origin and updates the files before building")
    parser.add_argument('--no-zip', action='store_false', dest='zip', help="Do not generate the zip files in the build output")
    parser.add_argument('--no-json', action='store_false', dest='json', help="Do not generate the json file in the build output")
    args = parser.parse_args()
    check_call(["git", "checkout", "-q", VERSION_TO_BRANCH[args.version], '--', 'plugins'])
    dest_dir = os.path.abspath(os.path.join(args.build_dir, args.version or ''))
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    if args.pull:
        check_call(["git", "pull", "-q"])
    if args.json:
        build_json(dest_dir)
    if args.zip:
        zip_files(dest_dir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Fix Track Numbers plugin for MusicBrainz Picard
# Copyright (C) 2017 Jonathan Bradley Whited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

PLUGIN_NAME = 'Fix Track Numbers'
PLUGIN_AUTHOR = 'Jonathan Bradley Whited'
PLUGIN_DESCRIPTION = '''
Fix the track numbers in a cluster by either using the track titles (1) or sequential order (2).

<ol>
  <li>
    The title should contain something like "#-#" (number dash number) and be unique.<br />
    All non-numbers and non-dashes will be removed when comparing the titles.<br />
    This is especially useful for Language Audio Lessons, like this:
    <pre>- Title: "Unit 1 - Lesson 10"</pre>
    For example, take the following titles and track numbers:
    <pre>
- Title: "Unit 1 - Lesson 1"  - Track #1
- Title: "Unit 1 - Lesson 2"  - Track #1
- Title: "Unit 2 - Lesson 10" - Track #2
- Title: "Unit 2 - Lesson 1"  - Track #2
</pre>
    The track numbers will be changed to:  1, 2, 4, 3<br />
    The 3rd one will be changed to Track #4 because Lesson 1 &lt; Lesson 10.<br />
    The titles will remain unchanged.
  </li>

  <li>The track numbers will be set based on the sequential order they appear within the cluster.</li>
</ol>

How to use:
<ol>
  <li>Cluster a group of files</li>
  <li>Right click on the cluster</li>
  <li>
    Then click one:
    <ul>
      <li>Plugins => Fix track numbers using titles</li>
      <li>Plugins => Fix track numbers using sequence</li>
    </ul>
  </li>
</ol>
'''
PLUGIN_VERSION = '0.2.1'
PLUGIN_API_VERSIONS = ['0.15', '1.0', '2.0']
PLUGIN_LICENSE = 'GPL-3.0-or-later'
PLUGIN_LICENSE_URL = 'https://www.gnu.org/licenses/gpl.txt'

from picard import log
from picard.cluster import Cluster
from picard.ui.itemviews import BaseAction, register_cluster_action

import re

# FIXME: for Python 3.0 and Picard 2.0, use String.format(...) instead of %?


class FixedTrack:

    def __init__(self, tracknumber=0, title=None, title_num1=0, title_num2=0):
        self.title = title
        self.title_num1 = title_num1
        self.title_num2 = title_num2
        self.tracknumber = tracknumber


class FixTrackNumsUsingTitles(BaseAction):
    NAME = 'Fix track numbers using titles'
    TITLE_REGEX = re.compile(r"[^\d\-]+")  # Only digits and '-' allowed

    def callback(self, objs):
        log.debug('[FixTrackNumsUsingTitles]')

        for cluster in objs:
            if not isinstance(cluster, Cluster) or not cluster.files:
                continue

            tracks = []  # Sorted list of FixedTrack

            for i, f in enumerate(cluster.files):
                if not f or not f.metadata or 'title' not in f.metadata:
                    log.debug('No file/metadata/title for [%i]' % (i))
                    continue

                track = FixedTrack(i + 1, f.metadata['title'])

                if not track.title:
                    log.debug('No title for [%i]' % (i))
                    continue

                title_nums = self.TITLE_REGEX.sub('', track.title)
                dash_index = title_nums.find('-')

                if dash_index < 0:
                    log.debug('No dash in [%i][%s]' % (i, title_nums))
                    continue

                try:
                    track.title_num1 = int(title_nums[:dash_index])
                    track.title_num2 = int(title_nums[dash_index + 1:])
                except ValueError:
                    log.debug('Invalid ints in [%i][%s]' % (i, title_nums))
                    continue

                was_added = False

                # Not empty?
                if tracks:
                    # Justin Timberlake?
                    for j, t in enumerate(tracks):
                        if was_added:
                            # Increment all track numbers above last added one
                            t.tracknumber += 1
                        # Don't do "<=" on title_num2 to preserve sequence
                        # - Case 1: "2-10" < "3-1"
                        # - Case 2: "2-1"  < "2-10"
                        elif ((track.title_num1 < t.title_num1) or
                              (track.title_num1 == t.title_num1 and
                               track.title_num2 < t.title_num2)):
                            # t.tracknumber will be updated in next loop cycle
                            track.tracknumber = t.tracknumber
                            tracks.insert(j, track)
                            was_added = True  # Don't break

                if not was_added:
                    tracks.append(track)

            # Let's build a dictionary of the new (fixed) track numbers
            new_tracks = {}

            for i, t in enumerate(tracks):
                # Assume title is unique
                new_tracks[t.title] = str(t.tracknumber)

            for i, f in enumerate(cluster.files):
                if not f or not f.metadata or 'title' not in f.metadata:
                    # Already logged
                    continue

                key = f.metadata['title']

                if not key:
                    # Already logged
                    continue
                if key not in new_tracks:
                    log.debug('No new track for [%i][%s]' % (i, key))
                    continue

                new_track = new_tracks[key]

                log.debug('Change [%s]=>[%s]' % (key, new_track))

                f.metadata['tracknumber'] = new_track
                f.metadata.changed = True
                f.update(signal=True)

            cluster.update()


class FixTrackNumsUsingSeq(BaseAction):
    NAME = 'Fix track numbers using sequence'

    def callback(self, objs):
        log.debug('[FixTrackNumsUsingSeq]')

        for cluster in objs:
            if not isinstance(cluster, Cluster) or not cluster.files:
                continue

            for i, f in enumerate(cluster.files):
                if not f or not f.metadata:
                    log.debug('No file/metadata for [%i]' % (i))
                    continue

                new_track = str(i + 1)

                if 'title' in f.metadata and f.metadata['title']:
                    log.debug('Change [%s]=>[%s]' %
                              (f.metadata['title'], new_track))

                f.metadata['tracknumber'] = new_track
                f.metadata.changed = True
                f.update(signal=True)

            cluster.update()

register_cluster_action(FixTrackNumsUsingTitles())
register_cluster_action(FixTrackNumsUsingSeq())



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'options_enhanced_titles.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PyQt5.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PyQt5.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PyQt5.QtWidgets import (QApplication, QCheckBox, QFrame, QGroupBox,
    QLabel, QScrollArea, QSizePolicy, QSpacerItem,
    QVBoxLayout, QWidget)

class Ui_EnhancedTitlesOptions(object):
    def setupUi(self, EnhancedTitlesOptions):
        if not EnhancedTitlesOptions.objectName():
            EnhancedTitlesOptions.setObjectName(u"EnhancedTitlesOptions")
        EnhancedTitlesOptions.setMinimumSize(QSize(100, 0))
        self.verticalLayout = QVBoxLayout(EnhancedTitlesOptions)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.scrollArea = QScrollArea(EnhancedTitlesOptions)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setFrameShape(QFrame.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 379, 502))
        self.verticalLayout_2 = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.titles_all_caps = QGroupBox(self.scrollAreaWidgetContents)
        self.titles_all_caps.setObjectName(u"titles_all_caps")
        self.verticalLayout_6 = QVBoxLayout(self.titles_all_caps)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.label_4 = QLabel(self.titles_all_caps)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setMouseTracking(True)
        self.label_4.setWordWrap(True)

        self.verticalLayout_6.addWidget(self.label_4)

        self.check_allcaps = QCheckBox(self.titles_all_caps)
        self.check_allcaps.setObjectName(u"check_allcaps")

        self.verticalLayout_6.addWidget(self.check_allcaps)


        self.verticalLayout_2.addWidget(self.titles_all_caps)

        self.check_for_sort_names = QGroupBox(self.scrollAreaWidgetContents)
        self.check_for_sort_names.setObjectName(u"check_for_sort_names")
        self.verticalLayout_3 = QVBoxLayout(self.check_for_sort_names)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.label = QLabel(self.check_for_sort_names)
        self.label.setObjectName(u"label")
        self.label.setWordWrap(True)

        self.verticalLayout_3.addWidget(self.label)

        self.check_tagging = QCheckBox(self.check_for_sort_names)
        self.check_tagging.setObjectName(u"check_tagging")
        self.check_tagging.setChecked(True)

        self.verticalLayout_3.addWidget(self.check_tagging)

        self.label_2 = QLabel(self.check_for_sort_names)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setWordWrap(True)

        self.verticalLayout_3.addWidget(self.label_2)

        self.check_album_aliases = QCheckBox(self.check_for_sort_names)
        self.check_album_aliases.setObjectName(u"check_album_aliases")
        self.check_album_aliases.setCheckable(True)
        self.check_album_aliases.setChecked(True)

        self.verticalLayout_3.addWidget(self.check_album_aliases)

        self.check_track_aliases = QCheckBox(self.check_for_sort_names)
        self.check_track_aliases.setObjectName(u"check_track_aliases")

        self.verticalLayout_3.addWidget(self.check_track_aliases)


        self.verticalLayout_2.addWidget(self.check_for_sort_names)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.verticalLayout.addWidget(self.scrollArea)


        self.retranslateUi(EnhancedTitlesOptions)

        QMetaObject.connectSlotsByName(EnhancedTitlesOptions)
    # setupUi

    def retranslateUi(self, EnhancedTitlesOptions):
        EnhancedTitlesOptions.setWindowTitle(QCoreApplication.translate("EnhancedTitlesOptions", u"Form", None))
        self.titles_all_caps.setTitle(QCoreApplication.translate("EnhancedTitlesOptions", u"Titles in All Caps", None))
        self.label_4.setText(QCoreApplication.translate("EnhancedTitlesOptions", u"If this option is enabled, text in all caps will not be affected by the $title_lang script function.", None))
        self.check_allcaps.setText(QCoreApplication.translate("EnhancedTitlesOptions", u"Ignore titles in all caps.", None))
        self.check_for_sort_names.setTitle(QCoreApplication.translate("EnhancedTitlesOptions", u"Tagging", None))
        self.label.setText(QCoreApplication.translate("EnhancedTitlesOptions", u"<html><head/><body><p>If this option is enabled, the fields albumsort and titlesort will be filled. If you are only interested in the scripting functions, then disable this to reduce the time it takes to load releases.</p></body></html>", None))
        self.check_tagging.setText(QCoreApplication.translate("EnhancedTitlesOptions", u"Tag albumsort and titlesort fields", None))
        self.label_2.setText(QCoreApplication.translate("EnhancedTitlesOptions", u"<html><head/><body><p>If this option is enabled, the plugin will first check if there are any sort names already available in MusicBrainz. Otherwise, it will swap the title's prefix directly. Enabling this option will increase the time it takes to load releases, especially for tracks.</p></body></html>", None))
        self.check_album_aliases.setText(QCoreApplication.translate("EnhancedTitlesOptions", u"Check for album aliases", None))
        self.check_track_aliases.setText(QCoreApplication.translate("EnhancedTitlesOptions", u"Check for track aliases", None))
    # retranslateUi




PLUGIN_NAME = "Deezer cover art"
PLUGIN_AUTHOR = "Fabio Forni <livingsilver94>"
PLUGIN_DESCRIPTION = "Fetch cover arts from Deezer"
PLUGIN_VERSION = '1.2.2'
PLUGIN_API_VERSIONS = ['2.5']
PLUGIN_LICENSE = "GPL-3.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-3.0.html"

from typing import Any, List, Optional
from urllib.parse import urlsplit

import picard
from picard import config
from picard.coverart import providers
from picard.coverart.image import CoverArtImage
from picard.util.astrcmp import astrcmp
from PyQt5 import QtNetwork as QtNet

from .deezer import Client, SearchOptions, obj
from .options import Ui_Form

__version__ = PLUGIN_VERSION

DEFAULT_SIMILARITY_THRESHOLD = 0.6


def is_similar(str1: str, str2: str, min_similarity: float = DEFAULT_SIMILARITY_THRESHOLD) -> bool:
    if str1 in str2:
        return True
    return astrcmp(str1, str2) >= min_similarity


def is_deezer_url(url: str) -> bool:
    return 'deezer.com' in urlsplit(url).netloc


class OptionsPage(providers.ProviderOptions):
    NAME = 'Deezer'
    TITLE = 'Deezer'
    options = [
        config.TextOption('setting', 'deezerart_size', obj.CoverSize.BIG.value),
        config.FloatOption('setting', 'deezerart_min_similarity', DEFAULT_SIMILARITY_THRESHOLD),
    ]
    _options_ui = Ui_Form

    def load(self):
        for s in obj.CoverSize:
            self.ui.size.addItem(str(s.name).title(), userData=s.value)
        self.ui.size.setCurrentIndex(self.ui.size.findData(config.setting['deezerart_size']))
        self.ui.min_similarity.setValue(int(config.setting["deezerart_min_similarity"] * 100))

    def save(self):
        config.setting['deezerart_size'] = self.ui.size.currentData()
        config.setting['deezerart_min_similarity'] = float(self.ui.min_similarity.value()) / 100.0


class Provider(providers.CoverArtProvider):
    NAME = 'Deezer'
    OPTIONS = OptionsPage
    _log_prefix = 'Deezerart: '

    def __init__(self, coverart):
        super().__init__(coverart)
        self.client = Client(self.album.tagger.webservice)
        self._has_url_relation = False
        self._retry_search = False

    # Override.
    def queue_images(self):
        self.match_url_relations(['free streaming'], self._url_callback)
        if not self._has_url_relation:
            if not self._retry_search:
                search_opts = SearchOptions(artist=self._artist(), album=self.metadata['album'])
            else:
                try:
                    track = self.release['media'][0]['tracks'][1]['title']
                except (IndexError, KeyError):
                    self.error('cannot find a track name to retry a search. No cover found')
                    return self.FINISHED
                else:
                    search_opts = SearchOptions(artist=self._artist(), track=track)
            self.client.advanced_search(search_opts, self._queue_from_search)
        self.album._requests += 1
        return self.WAIT

    # Override.
    def error(self, msg):
        super().error(self._log_prefix + msg)

    def log_debug(self, msg: Any, *args):
        picard.log.debug(self._log_prefix + msg, *args)

    def _url_callback(self, url: str):
        if is_deezer_url(url):
            self._has_url_relation = True
            self.client.obj_from_url(url, self._queue_from_url)

    def _queue_from_url(self, album: obj.APIObject, error: QtNet.QNetworkReply.NetworkError):
        self.album._requests -= 1
        try:
            if error:
                self.error('could not get Deezer API object: {}'.format(error))
                return
            if not isinstance(album, obj.Album):
                self.error('API object is not an album')
                return
            cover_url = album.cover_url(obj.CoverSize(config.setting['deezerart_size']))
            self.queue_put(CoverArtImage(cover_url))
            self.log_debug('queued cover using an URL relation')
        finally:
            self.next_in_queue()

    def _queue_from_search(self, results: List[obj.APIObject], error: Optional[QtNet.QNetworkReply.NetworkError]):
        self.album._requests -= 1
        try:
            if error:
                self.error('could not fetch search results: {}'.format(error))
                return
            if len(results) == 0:
                if self._retry_search:
                    self.error('no results found')
                    return
                self._retry_search = True
                self.queue_images()
                return
            artist = self._artist()
            album = self.metadata['album']
            min_similarity = config.setting['deezerart_min_similarity']
            for result in results:
                if not isinstance(result, obj.Track):
                    continue
                if not is_similar(artist, result.artist.name, min_similarity):
                    self.log_debug('artist similarity below threshold: %r ~ %r', artist, result.artist.name)
                    continue
                if not is_similar(album, result.album.title, min_similarity):
                    self.log_debug('album similarity below threshold: %r ~ %r', album, result.album.title)
                    continue
                cover_url = result.album.cover_url(obj.CoverSize(config.setting['deezerart_size']))
                self.queue_put(CoverArtImage(cover_url))
                self.log_debug('queued cover using a Deezer search')
                return
            self.error('no result matched the criteria')
        finally:
            self.next_in_queue()

    def _artist(self) -> str:
        # If there are many artists, we want to search
        # the album in Deezer with just one as keyword.
        # Deezerart may not specify all the artists
        # MusicBrainz does, or it may use different separators.
        return self.metadata.getraw('~albumartists')[0]


providers.register_cover_art_provider(Provider)



# -*- coding: utf-8 -*-

# Copyright (C) 2019 Alex Rustler <alex_rustler@rambler.ru>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from picard import metadata
from picard.script import register_script_function

PLUGIN_NAME = "Replace Forbidden Symbols"
PLUGIN_AUTHOR = "Alex Rustler <alex_rustler@rambler.ru>"
PLUGIN_VERSION = "0.3"
PLUGIN_API_VERSIONS = ["0.9", "0.10", "0.11", "0.15", "2.0", "2.2"]
PLUGIN_LICENSE = "GPL-3.0-or-later"
PLUGIN_LICENSE_URL = "https://gnu.org/licenses/gpl.html"
PLUGIN_DESCRIPTION = '''Replaces Windows forbidden symbols: :, /, *, ?, ", ., | etc.
                    with a similar UNICODE version.
                    Currently replaces characters on "album", "artist",
                    "title", "albumartist", "releasetype", "label" tags.
                    Also add $replace_forbidden() function for Tagger.
                    Example: $set(composer,$script_forbidden(%composer%))
'''

CHAR_TABLE = {

    # forbidden
    ":": "∶",
    "/": "⁄",
    "*": "∗",
    "?": "？",
    '"': '″',
    '\\': '⧵',
    '.': '․',
    '|': 'ǀ',
    '<': '‹',
    '>': '›'
}

FILTER_TAGS = [
    "album",
    "artist",
    "title",
    "albumartist",
    "releasetype",
    "label",
]


def sanitize(char):
    if char in CHAR_TABLE:
        return CHAR_TABLE[char]
    return char


def fix_forbidden(word):
    return "".join(sanitize(char) for char in word)


def replace_forbidden(value):
    return [fix_forbidden(x) for x in value]


def script_replace_forbidden(parser, value):
    return fix_forbidden(value)


def main(tagger, metadata, *args):
    for name, value in metadata.rawitems():
        if name in FILTER_TAGS:
            metadata[name] = replace_forbidden(value)


metadata.register_track_metadata_processor(main)
metadata.register_album_metadata_processor(main)

register_script_function(script_replace_forbidden, name="replace_forbidden")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2023 Bob Swift (rdswift)
# Copyright (C) 2023 Ruud van Asseldonk (ruuda)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'Additional Artists Variables'
PLUGIN_AUTHOR = 'Bob Swift (rdswift)'
PLUGIN_DESCRIPTION = '''
This plugin provides specialized album and track variables for use in
naming scripts. It is based on the "Album Artist Extension" plugin, but
expands the functionality to also include track artists. Note that it
cannot be used as a direct drop-in replacement for the "Album Artist
Extension" plugin because the variables are provided with different
names.  This will require changes to existing scripts if switching to
this plugin.
<br /><br />
Please see the <a href="https://github.com/rdswift/picard-plugins/blob/2.0_RDS_Plugins/plugins/additional_artists_variables/docs/README.md">user guide</a> on GitHub for more information.
'''
PLUGIN_VERSION = '1.0'
PLUGIN_API_VERSIONS = ['2.0', '2.1', '2.2', '2.7', '2.9', '2.10', '2.11']
PLUGIN_LICENSE = 'GPL-2.0-or-later'
PLUGIN_LICENSE_URL = 'https://www.gnu.org/licenses/gpl-2.0.html'

PLUGIN_USER_GUIDE_URL = 'https://github.com/rdswift/picard-plugins/blob/2.0_RDS_Plugins/plugins/additional_artists_variables/docs/README.md'

from operator import itemgetter

from picard import config, log
from picard.metadata import (register_album_metadata_processor,
                             register_track_metadata_processor)
from picard.plugin import PluginPriority


ID_ALIASES_ARTIST_NAME = '894afba6-2816-3c24-8072-eadb66bd04bc'
ID_ALIASES_LEGAL_NAME = 'd4dcd0c0-b341-3612-a332-c0ce797b25cf'


def process_artists(album_id, source_metadata, destination_metadata, source_type):
    # Test for valid metadata node.
    # The 'artist-credit' key should always be there.
    # This check is to avoid a runtime error if it doesn't exist for some reason.
    if 'artist-credit' in source_metadata:
        # Initialize variables to default values
        sort_pri_artist = ''
        sort_pri_artist_cred = ''
        std_artist = ''
        cred_artist = ''
        sort_artist = ''
        cred_sort_artist = ''
        legal_artist = ''
        additional_std_artist = ''
        additional_cred_artist = ''
        additional_sort_artist = ''
        additional_cred_sort_artist = ''
        additional_legal_artist = ''
        std_artist_list = []
        cred_artist_list = []
        sort_artist_list = []
        cred_sort_artist_list = []
        legal_artist_list = []
        artist_count = 0
        artist_ids = []
        artist_types = []
        artist_join_phrases = []
        for artist_credit in source_metadata['artist-credit']:
            # Initialize temporary variables for each loop.
            temp_std_name = ''
            temp_sort_name = ''
            temp_cred_name = ''
            temp_cred_sort_name = ''
            temp_legal_name = ''
            temp_legal_sort_name = ''
            temp_phrase = ''
            temp_id = ''
            temp_type = ''
            # Check if there is a 'joinphrase' specified.
            if 'joinphrase' in artist_credit:
                temp_phrase = artist_credit['joinphrase']
            else:
                metadata_error(album_id, 'artist-credit.joinphrase', source_type)
            # Check if there is a 'name' specified.
            if 'name' in artist_credit:
                temp_cred_name = artist_credit['name']
            else:
                metadata_error(album_id, 'artist-credit.name', source_type)
            # Check if there is an 'artist' specified.
            if 'artist' in artist_credit:
                if 'id' in artist_credit['artist']:
                    temp_id = artist_credit['artist']['id']
                else:
                    metadata_error(album_id, 'artist-credit.artist.id', source_type)
                if 'name' in artist_credit['artist']:
                    temp_std_name = artist_credit['artist']['name']
                else:
                    metadata_error(album_id, 'artist-credit.artist.name', source_type)
                if 'sort-name' in artist_credit['artist']:
                    temp_sort_name = artist_credit['artist']['sort-name']
                    temp_cred_sort_name = temp_sort_name
                else:
                    metadata_error(album_id, 'artist-credit.artist.sort-name', source_type)
                if 'type' in artist_credit['artist']:
                    temp_type = artist_credit['artist']['type']
                else:
                    metadata_error(album_id, 'artist-credit.artist.type', source_type)
                if 'aliases' in artist_credit['artist']:
                    for item in artist_credit['artist']['aliases']:
                        if 'type-id' in item and item['type-id'] == ID_ALIASES_LEGAL_NAME:
                            if 'ended' in item and not item['ended']:
                                if 'name' in item:
                                    temp_legal_name = item['name']
                                if 'sort-name' in item:
                                    temp_legal_sort_name = item['sort-name']
                        if 'type-id' in item and item['type-id'] == ID_ALIASES_ARTIST_NAME:
                            if 'name' in item and 'sort-name' in item and str(item['name']).lower() == temp_cred_name.lower():
                                temp_cred_sort_name = item['sort-name']
                tag_list = []
                if config.setting['max_genres']:
                    for tag_type in ['user-genres', 'genres', 'user-tags', 'tags']:
                        if tag_type in artist_credit['artist']:
                            for item in sorted(sorted(artist_credit['artist'][tag_type], key=itemgetter('name')), key=itemgetter('count'), reverse=True):
                                if item['count'] > 0:
                                    tag_list.append(item['name'])
                    tag_list = tag_list[:config.setting['max_genres']]
            else:
                # No 'artist' specified.  Log as an error.
                metadata_error(album_id, 'artist-credit.artist', source_type)
            std_artist += temp_std_name + temp_phrase
            cred_artist += temp_cred_name + temp_phrase
            sort_artist += temp_sort_name + temp_phrase
            cred_sort_artist += temp_cred_sort_name + temp_phrase
            artist_types.append(temp_type if temp_type else 'unknown',)
            artist_join_phrases.append(temp_phrase if temp_phrase else '\u200B',)
            if temp_legal_name:
                legal_artist += temp_legal_name + temp_phrase
                legal_artist_list.append(temp_legal_name,)
            else:
                # Use standardized name for combined string if legal name not available
                legal_artist += temp_std_name + temp_phrase
                # Use 'n/a' for list if legal name not available
                legal_artist_list.append('n/a',)
            if temp_std_name:
                std_artist_list.append(temp_std_name,)
            if temp_sort_name:
                sort_artist_list.append(temp_sort_name,)
            if temp_cred_name:
                cred_artist_list.append(temp_cred_name,)
            if temp_cred_sort_name:
                cred_sort_artist_list.append(temp_cred_sort_name,)
            if temp_id:
                artist_ids.append(temp_id,)
            if artist_count < 1:
                if temp_id:
                    destination_metadata['~artists_{0}_primary_id'.format(source_type,)] = temp_id
                destination_metadata['~artists_{0}_primary_std'.format(source_type,)] = temp_std_name
                destination_metadata['~artists_{0}_primary_cred'.format(source_type,)] = temp_cred_name
                destination_metadata['~artists_{0}_primary_sort'.format(source_type,)] = temp_sort_name
                destination_metadata['~artists_{0}_primary_cred_sort'.format(source_type,)] = temp_cred_sort_name
                destination_metadata['~artists_{0}_primary_legal'.format(source_type,)] = temp_legal_name
                destination_metadata['~artists_{0}_primary_sort_legal'.format(source_type,)] = temp_legal_sort_name
                sort_pri_artist += temp_sort_name + temp_phrase
                sort_pri_artist_cred += temp_cred_sort_name + temp_phrase
                if tag_list and source_type == 'album':
                    destination_metadata['~artists_{0}_primary_tags'.format(source_type,)] = tag_list
            else:
                sort_pri_artist += temp_std_name + temp_phrase
                additional_std_artist += temp_std_name + temp_phrase
                additional_cred_artist += temp_cred_name + temp_phrase
                additional_sort_artist += temp_sort_name + temp_phrase
                if temp_legal_name:
                    additional_legal_artist += temp_legal_name + temp_phrase
                else:
                    additional_legal_artist += temp_std_name + temp_phrase
                if temp_cred_sort_name:
                    additional_cred_sort_artist += temp_cred_sort_name + temp_phrase
                    sort_pri_artist_cred += temp_cred_sort_name + temp_phrase
                else:
                    additional_cred_sort_artist += temp_sort_name + temp_phrase
                    sort_pri_artist_cred += temp_sort_name + temp_phrase
            artist_count += 1
    else:
        # No valid metadata found.  Log as error.
        metadata_error(album_id, 'artist-credit', source_type)
    additional_std_artist_list = std_artist_list[1:]
    additional_cred_artist_list = cred_artist_list[1:]
    additional_sort_artist_list = sort_artist_list[1:]
    additional_cred_sort_artist_list = cred_sort_artist_list[1:]
    additional_legal_artist_list = legal_artist_list[1:]
    additional_artist_ids = artist_ids[1:]
    if additional_artist_ids:
        destination_metadata['~artists_{0}_additional_id'.format(source_type,)] = additional_artist_ids
    if additional_std_artist:
        destination_metadata['~artists_{0}_additional_std'.format(source_type,)] = additional_std_artist
    if additional_cred_artist:
        destination_metadata['~artists_{0}_additional_cred'.format(source_type,)] = additional_cred_artist
    if additional_sort_artist:
        destination_metadata['~artists_{0}_additional_sort'.format(source_type,)] = additional_sort_artist
    if additional_cred_sort_artist:
        destination_metadata['~artists_{0}_additional_cred_sort'.format(source_type,)] = additional_cred_sort_artist
    if additional_legal_artist:
        destination_metadata['~artists_{0}_additional_legal'.format(source_type,)] = additional_legal_artist
    if additional_std_artist_list:
        destination_metadata['~artists_{0}_additional_std_multi'.format(source_type,)] = additional_std_artist_list
    if additional_cred_artist_list:
        destination_metadata['~artists_{0}_additional_cred_multi'.format(source_type,)] = additional_cred_artist_list
    if additional_sort_artist_list:
        destination_metadata['~artists_{0}_additional_sort_multi'.format(source_type,)] = additional_sort_artist_list
    if additional_cred_sort_artist_list:
        destination_metadata['~artists_{0}_additional_cred_sort_multi'.format(source_type,)] = additional_cred_sort_artist_list
    if additional_legal_artist_list:
        destination_metadata['~artists_{0}_additional_legal_multi'.format(source_type,)] = additional_legal_artist_list
    if std_artist:
        destination_metadata['~artists_{0}_all_std'.format(source_type,)] = std_artist
    if cred_artist:
        destination_metadata['~artists_{0}_all_cred'.format(source_type,)] = cred_artist
    if cred_sort_artist:
        destination_metadata['~artists_{0}_all_cred_sort'.format(source_type,)] = cred_sort_artist
    if sort_artist:
        destination_metadata['~artists_{0}_all_sort'.format(source_type,)] = sort_artist
    if legal_artist:
        destination_metadata['~artists_{0}_all_legal'.format(source_type,)] = legal_artist
    if std_artist_list:
        destination_metadata['~artists_{0}_all_std_multi'.format(source_type,)] = std_artist_list
    if cred_artist_list:
        destination_metadata['~artists_{0}_all_cred_multi'.format(source_type,)] = cred_artist_list
    if sort_artist_list:
        destination_metadata['~artists_{0}_all_sort_multi'.format(source_type,)] = sort_artist_list
    if cred_sort_artist_list:
        destination_metadata['~artists_{0}_all_cred_sort_multi'.format(source_type,)] = cred_sort_artist_list
    if legal_artist_list:
        destination_metadata['~artists_{0}_all_legal_multi'.format(source_type,)] = legal_artist_list
    if sort_pri_artist:
        destination_metadata['~artists_{0}_all_sort_primary'.format(source_type,)] = sort_pri_artist
    if artist_types:
        destination_metadata['~artists_{0}_all_types'.format(source_type,)] = artist_types
    if artist_join_phrases:
        destination_metadata['~artists_{0}_all_join_phrases'.format(source_type,)] = artist_join_phrases
    if artist_count:
        destination_metadata['~artists_{0}_all_count'.format(source_type,)] = artist_count


def make_album_vars(album, album_metadata, release_metadata):
    album_id = release_metadata['id'] if release_metadata else 'No Album ID'
    process_artists(album_id, release_metadata, album_metadata, 'album')


def make_track_vars(album, album_metadata, track_metadata, release_metadata):
    album_id = release_metadata['id'] if release_metadata else 'No Album ID'
    process_artists(album_id, track_metadata, album_metadata, 'track')


def metadata_error(album_id, metadata_element, metadata_group):
    log.error("{0}: {1!r}: Missing '{2}' in {3} metadata.".format(
        PLUGIN_NAME, album_id, metadata_element, metadata_group,))


# Register the plugin to run at a LOW priority so that other plugins that
# modify the artist information can complete their processing and this plugin
# is working with the latest updated data.
register_album_metadata_processor(make_album_vars, priority=PluginPriority.LOW)
register_track_metadata_processor(make_track_vars, priority=PluginPriority.LOW)



# MusicBrainz Picard plugin to re-order sides of a release.
# Copyright (C) 2016  David Mandelberg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

PLUGIN_NAME = 'Re-order sides of a release'
PLUGIN_AUTHOR = 'David Mandelberg, Sambhav Kothari'
PLUGIN_DESCRIPTION = """\
  Split mediums and re-order sides to match side order rather than
  medium order. E.g., if a release has two mediums with track numbers
  <em>A1, A2, ..., D1, D2, ...</em> and <em>B1, B2, ..., C1, C2,
  ...</em>, this plugin will split the release into four mediums and
  reorder the new mediums so that the track numbers are <em>A1, A2,
  ..., B1, B2, ..., C1, C2, ..., D1, D2, ...</em>

  This is primarily intended to make vinyl records designed for record
  changers
  (https://en.wikipedia.org/wiki/Record_changer#Automatic_sequencing)
  play in the correct order."""
PLUGIN_VERSION = '1.2'
PLUGIN_API_VERSIONS = ['2.0']
PLUGIN_LICENSE = 'GPL-3.0-or-later'
PLUGIN_LICENSE_URL = 'https://www.gnu.org/licenses/gpl-3.0.html'

import collections

from picard.metadata import \
  register_album_metadata_processor, register_track_metadata_processor


# List of strings that represent the medium side of a track when the
# track number is one of these strings followed by digits. The order of
# this list specifies the order of the sides.
SIDE_PREFIXES = [
  'A',
  'B',
  'C',
  'D',
  'E',
  'F',
  'G',
  'H',
  'I',
  'J',
  'K',
  'L',
  'M',
  'N',
  'O',
  'P',
  'Q',
  'R',
  'S',
  'T',
  'U',
  'V',
  'W',
  'X',
  'Y',
  'Z',
  ]

# Map from release MBID to information that we can use to re-group and
# re-order the tracks. The per-release info (value of this map) is an
# OrderedDict mapping side name to a sequence of (discnumber of the
# side, first tracknumber of the side, last tracknumber of the side).
# The order of the OrderedDict specifies the intended order of the
# sides.
release_to_side_info = {}

def tracknumber_to_side(tracknumber):
  """Given a track "number", return the side, or None"""

  for side_prefix in SIDE_PREFIXES:
    if not tracknumber.startswith(side_prefix):
      continue

    number_in_side = tracknumber[len(side_prefix):]
    try:
      int(number_in_side)
    except ValueError:
      continue

    return side_prefix

  return None

def get_side_info(release):
  """Return side info (see release_to_side_info), or None if we don't
  want to reorder the sides."""

  side_info = collections.OrderedDict()

  for medium in release['media']:
    current_side = None

    for track in medium['tracks']:
      tracknumber = track['number']
      trackside = tracknumber_to_side(tracknumber)

      try:
        int_tracknumber = int(track['position'])
      except ValueError:
        # Non-integer tracknumber, so give up.
        return None

      if trackside is None:
        # If any track has no side information, we don't reorder
        # anything.
        return None

      if current_side is not None and current_side == trackside:
        # Another track of the same side as before, so just update the
        # last tracknumber for this side.
        if int_tracknumber > side_info[current_side][2]:
          side_info[current_side][2] = int_tracknumber
        continue

      # At this point, we're on the first track of a new side.

      current_side = trackside

      if current_side in side_info:
        # We've already seen this side somewhere else, so give up.
        return None

      try:
        side_info[current_side] = [
          int(medium['position']),
          int_tracknumber,
          int_tracknumber,
          ]
      except ValueError:
        # Non-integer position, so give up.
        return None

  # At this point, we know that all tracknumbers include side
  # information, and no sides appear in more than one place (every
  # side is contiguous).

  sides_in_order_of_appearance = list(side_info.keys())
  sides_in_intended_order = sorted(
    sides_in_order_of_appearance,
    key=lambda s: SIDE_PREFIXES.index(s),
    )
  if sides_in_order_of_appearance == sides_in_intended_order:
    # Sides are already in the right order, so we don't need to do
    # anything.
    return None

  # Re-order side_info to match the intended order.
  side_info_ordered = collections.OrderedDict()
  for side in sides_in_intended_order:
    side_info_ordered[side] = side_info[side]
  side_info = side_info_ordered

  return side_info

def find_side(side_info, metadata):
  """Given side info and track metadata, return the side that the
  track belongs to."""

  orig_discnumber = int(metadata['discnumber'])
  orig_tracknumber = int(metadata['tracknumber'])

  for side_item in side_info.items():
    (
      side,
      (
        side_discnumber,
        side_first_tracknumber,
        side_last_tracknumber,
        ),
      ) = side_item

    if side_discnumber == orig_discnumber \
        and side_first_tracknumber <= orig_tracknumber \
        and orig_tracknumber <= side_last_tracknumber:
      return side

  raise RuntimeError('Unable to find side for track: ' + metadata)

def analyze_release(tagger, metadata, release):
  """Analyze a release to determine if its sides should be
  re-ordered"""

  side_info = get_side_info(release)
  if side_info is not None:
    release_to_side_info[metadata['musicbrainz_albumid']] = side_info

register_album_metadata_processor(analyze_release)

def reorder_sides(tagger, metadata, *args):
  """Re-order sides of a release, using the information in
  release_to_side_info."""

  if metadata['musicbrainz_albumid'] not in release_to_side_info:
    return

  side_info = release_to_side_info[metadata['musicbrainz_albumid']]
  all_sides = list(side_info.keys())

  side = find_side(side_info, metadata)
  side_first_tracknumber = side_info[side][1]
  side_last_tracknumber = side_info[side][2]

  metadata['totaldiscs'] = str(len(all_sides))
  metadata['discnumber'] = str(all_sides.index(side) + 1)

  metadata['totaltracks'] = \
    str(side_last_tracknumber - side_first_tracknumber + 1)
  metadata['tracknumber'] = \
    str(int(metadata['tracknumber']) - side_first_tracknumber + 1)

register_track_metadata_processor(reorder_sides)



PLUGIN_NAME = 'Remove Perfect Albums'
PLUGIN_AUTHOR = 'ichneumon, hrglgrmpf'
PLUGIN_DESCRIPTION = '''Remove all perfectly matched albums from the selection.'''
PLUGIN_VERSION = '0.3'
PLUGIN_API_VERSIONS = ['2.0', '2.1', '2.2', '2.3']
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from PyQt5.QtCore import QCoreApplication

from picard.album import Album
from picard.ui.itemviews import BaseAction, register_album_action


class RemovePerfectAlbums(BaseAction):
    NAME = 'Remove perfect albums'

    def callback(self, objs):
        for album in objs:
            if (isinstance(album, Album) and album.loaded
               and album.is_complete() and album.get_num_unsaved_files() == 0):
                self.tagger.remove_album(album)
            QCoreApplication.processEvents()


register_album_action(RemovePerfectAlbums())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021, 2023 Bob Swift (rdswift)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'Submit ISRC'
PLUGIN_AUTHOR = 'Bob Swift'
PLUGIN_DESCRIPTION = '''
<p>
Adds a right click option on an album to submit the ISRCs to the MusicBrainz server
specified in the Options.
</p><p>
To use this function, you must first match your files to the appropriate tracks for
a release.  Once this is done, but before you save your files if you have Picard set
to overwrite the 'isrc' tag in your files, right-click the release and select "Submit
ISRCs" in the "Plugins" section.  For each file that has a single valid ISRC in its
metadata, the ISRC will be added to the recording on the release if it does not
already exist.  Once all tracks for the release have been processed, the missing
ISRCs will be submitted to MusicBrainz.
</p><p>
If a file's metadata contains multiple ISRCs, such as if the file has already been
tagged, then no ISRCs will be submitted for that file.
</p><p>
If one of the files contains an invalid ISRC, or if the same ISRC appears in the
metadata for two or more files, then a notice will be displayed and the submission
process will be aborted.
</p><p>
When ISRCs have been submitted, a notice will be displayed showing whether or not
the submission was successful.
</p>
'''
PLUGIN_VERSION = '1.1'
PLUGIN_API_VERSIONS = ['2.0', '2.1', '2.2', '2.3', '2.6', '2.9']
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.txt"

import re

from picard import log, PICARD_VERSION
from picard.ui.itemviews import BaseAction, register_album_action
from picard.version import Version
from picard.webservice.api_helpers import MBAPIHelper, _wrap_xml_metadata
from PyQt5 import QtCore, QtWidgets

RE_VALIDATE_ISRC = re.compile(r'^[A-Z]{2}[A-Z0-9]{3}[0-9]{7}$')

NEW_MBAPIHelper = (PICARD_VERSION >= Version(2, 9, 0, 'beta', 2))

XML_HEADER = '<recording-list>'
XML_TEMPLATE = '<recording id="{0}"><isrc-list count="1"><isrc id="{1}" /></isrc-list></recording>'
XML_FOOTER = '</recording-list>'

Q_ERROR_CODES = {
    0: 'No error',
    1: "The remote server refused the connection (the server is not accepting requests).",
    2: "The remote server closed the connection prematurely, before the entire reply was received and processed.",
    3: "The remote host name was not found (invalid hostname).",
    4: "The connection to the remote server timed out.",
    5: "The operation was canceled via calls to abort() or close() before it was finished.",
    6: "The SSL/TLS handshake failed and the encrypted channel could not be established. The sslErrors() signal should have been emitted.",
    7: "The connection was broken due to disconnection from the network, however the system has initiated roaming to another access point. The request should be resubmitted and will be processed as soon as the connection is re-established.",
    8: "The connection was broken due to disconnection from the network or failure to start the network.",
    9: "The background request is not currently allowed due to platform policy.",
    10: "While following redirects, the maximum limit was reached.",
    11: "While following redirects, the network access API detected a redirect from a encrypted protocol (https) to an unencrypted one (http).",
    99: "An unknown network-related error was detected.",
    101: "The connection to the proxy server was refused (the proxy server is not accepting requests).",
    102: "The proxy server closed the connection prematurely, before the entire reply was received and processed.",
    103: "The proxy host name was not found (invalid proxy hostname).",
    104: "The connection to the proxy timed out or the proxy did not reply in time to the request sent.",
    105: "The proxy requires authentication in order to honour the request but did not accept any credentials offered (if any).",
    199: "An unknown proxy-related error was detected.",
    201: "The access to the remote content was denied (similar to HTTP error 403).",
    202: "The operation requested on the remote content is not permitted.",
    203: "The remote content was not found at the server (similar to HTTP error 404).",
    204: "The remote server requires authentication to serve the content but the credentials provided were not accepted (if any).",
    205: "The request needed to be sent again, but this failed for example because the upload data could not be read a second time.",
    206: "The request could not be completed due to a conflict with the current state of the resource.",
    207: "The requested resource is no longer available at the server.",
    299: "An unknown error related to the remote content was detected.",
    301: "The Network Access API cannot honor the request because the protocol is not known.",
    302: "The requested operation is invalid for this protocol.",
    399: "A breakdown in protocol was detected (parsing error, invalid or unexpected responses, etc.).",
    401: "The server encountered an unexpected condition which prevented it from fulfilling the request.",
    402: "The server does not support the functionality required to fulfill the request.",
    403: "The server is unable to handle the request at this time.",
    499: "An unknown error related to the server response was detected.",
}


def validate_isrc(isrc):
    """Verify that the provided ISRC matches the standard pattern for a valid ISRC.

    Args:
        isrc (str): ISRC to validate

    Returns:
        str: Properly formatted ISRC (upper case with no spaces or hyphens) if valid, otherwise None
    """
    formatted_isrc = str(isrc).upper().replace(' ', '').replace('-', '')
    if re.match(RE_VALIDATE_ISRC, formatted_isrc):
        return formatted_isrc
    return None


def show_popup(title, content, window=None):
    """Display a pop-up dialog.

    Args:
        title (str): Title for the pop-up dialog.
        content (str): Test to be displayed in the pop-up dialog..
        window (object, optional): Parent object for the dialog. Defaults to None.
    """
    QtWidgets.QMessageBox.information(
        window,
        title,
        content,
        QtWidgets.QMessageBox.Ok,
        QtWidgets.QMessageBox.Ok
    )


class SubmitAlbumISRCs(BaseAction):
    NAME = 'Submit ISRCs'

    def callback(self, album):
        if not album:
            log.error("{0}: No album specified for submitting ISRCs.".format(PLUGIN_NAME,))
            return

        log.info("{0}: Submitting ISRCs for: {1}".format(PLUGIN_NAME, album[0].metadata['album'],))
        if not album[0].tracks:
            log.debug("{0}: No tracks found in album: {1}".format(PLUGIN_NAME, album[0].metadata['album'],))
            show_popup('Error', 'No tracks found in the album.')
            return

        isrcs = {}
        multi_isrcs = []
        for track in album[0].tracks:
            if not track.files:
                continue
            audio_file = track.files[0]
            metadata = track.metadata
            file_metadata = audio_file.orig_metadata

            # No ISRC found in the file
            if 'isrc' not in file_metadata:
                continue

            # Get string of existing ISRCs on MusicBrainz
            if 'isrc' in metadata:
                mb_isrc = metadata['isrc'].upper()
            else:
                mb_isrc = ''

            # Get ISRC string from the file
            file_isrc = file_metadata['isrc']

            # Multiple ISRCs found in the file (don't process)
            if ';' in file_isrc:
                multi_isrcs.append('  {0} - {1}'.format(metadata['tracknumber'], metadata['title']))
                log.info("{0}: Multiple ISRCs found on track {1} (not processed): {2}".format(PLUGIN_NAME, metadata['tracknumber'], file_isrc))
                continue

            isrc = validate_isrc(file_isrc)

            # ISRC does not pass validation test
            if not isrc:
                log.debug("{0}: Invalid ISRC found on track {1}: {2}".format(PLUGIN_NAME, metadata['tracknumber'], file_isrc))
                show_popup('Error', "Invalid ISRC found on track {0}: '{1}'".format(metadata['tracknumber'], file_isrc))
                return

            # ISRC already found on another track for this album
            if isrc in isrcs:
                log.debug("{0}: Duplicate ISRC found on track {1}: {2}".format(PLUGIN_NAME, metadata['tracknumber'], file_isrc))
                show_popup('Error', "Duplicate ISRC found on track {0}: '{1}'".format(metadata['tracknumber'], file_isrc))
                return

            # ISRC already associated with that track (MusicBrainz recording)
            if isrc in mb_isrc:
                continue

            # New ISRC added for submission
            log.debug("{0}: Adding ISRC '{1}' for track {2} - \"{3}\"".format(PLUGIN_NAME, isrc, metadata['tracknumber'], metadata['title'],))
            isrcs[isrc] = metadata['musicbrainz_recordingid']

        if multi_isrcs:
            multiple_msg = '\n\nThe following track audio files contained multiple ISRCs (not submitted):\n' + '\n'.join(multi_isrcs)
        else:
            multiple_msg = ''

        # Save count of new ISRCs to display in success message
        self.isrc_count = len(isrcs)

        # Nothing to submit
        if not isrcs:
            log.debug("{0}: No new ISRCs found in album: {1}".format(PLUGIN_NAME, album[0].metadata['album'],))
            show_popup('Error', 'No new ISRCs found for the tracks in the album.{0}'.format(multiple_msg,))
            return

        if multiple_msg:
            show_popup('Submitting', 'Submitting {0} ISRC{1}.{2}'.format(self.isrc_count, '' if self.isrc_count == 1 else 's', multiple_msg,))

        # Build the xml data payload
        xml_items = [XML_HEADER]
        for isrc, recording in isrcs.items():
            xml_items.append(XML_TEMPLATE.format(recording, isrc))
        xml_items.append(XML_FOOTER)
        data = _wrap_xml_metadata(''.join(xml_items))

        # Initialize the MusicBrainz API Helper
        webservice = album[0].tagger.webservice
        helper = MBAPIHelper(webservice)

        # Set up parameters for the helper
        client_string = 'Picard_Plugin_{0}-v{1}'.format(PLUGIN_NAME, PLUGIN_VERSION).replace(' ', '_')
        handler = self.submission_handler
        path = '/recording' if NEW_MBAPIHelper else ['recording']
        params = {"client": client_string}

        return helper.post(path, data, handler, priority=True,
                         queryargs=params, parse_response_type="xml",
                         request_mimetype="application/xml; charset=utf-8")

    def submission_handler(self, document, reply, error):
        if not error:
            show_popup('Success', 'Successfully submitted {0} ISRC{1}.'.format(
                self.isrc_count,
                '' if self.isrc_count == 1 else 's',
            ))
            return

        # Decode response if necessary.
        xml_text = str(document, 'UTF-8') if isinstance(document, (bytes, bytearray, QtCore.QByteArray)) else str(document)

        # Build error text message from returned xml payload
        err_text = ''
        matches = re.findall(r'<text>(.*?)</text>', xml_text)
        if matches:
            err_text = '\n'.join(matches)
        else:
            err_text = ''

        # Use standard QNetworkReply error messages if no message was provided in the xml payload
        if not err_text:
            err_text = Q_ERROR_CODES[error] if error in Q_ERROR_CODES else 'There was no error message provided.'

        show_popup('Error', "There was an error processing the ISRC submission.  Please try again.\n\nError Code: {0}\n\n{1}".format(error, err_text))


register_album_action(SubmitAlbumISRCs())



# -*- coding: utf-8 -*-

PLUGIN_NAME = 'No release'
PLUGIN_AUTHOR = 'Johannes Weißl, Philipp Wolfer'
PLUGIN_DESCRIPTION = '''Do not store specific release information in releases of unknown origin.'''
PLUGIN_VERSION = '0.3'
PLUGIN_API_VERSIONS = ['2.0']

from PyQt5 import QtCore, QtGui, QtWidgets

from picard import config
from picard.album import Album
from picard.metadata import register_album_metadata_processor, register_track_metadata_processor
from picard.ui.options import register_options_page, OptionsPage
from picard.ui.itemviews import BaseAction, register_album_action
from picard.config import BoolOption, TextOption


class Ui_NoReleaseOptionsPage(object):

    def setupUi(self, NoReleaseOptionsPage):
        NoReleaseOptionsPage.setObjectName('NoReleaseOptionsPage')
        NoReleaseOptionsPage.resize(394, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(NoReleaseOptionsPage)
        self.verticalLayout.setObjectName('verticalLayout')
        self.groupBox = QtWidgets.QGroupBox(NoReleaseOptionsPage)
        self.groupBox.setObjectName('groupBox')
        self.vboxlayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.vboxlayout.setObjectName('vboxlayout')
        self.norelease_enable = QtWidgets.QCheckBox(self.groupBox)
        self.norelease_enable.setObjectName('norelease_enable')
        self.vboxlayout.addWidget(self.norelease_enable)
        self.label = QtWidgets.QLabel(self.groupBox)
        self.label.setObjectName('label')
        self.vboxlayout.addWidget(self.label)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName('horizontalLayout')
        self.norelease_strip_tags = QtWidgets.QLineEdit(self.groupBox)
        self.norelease_strip_tags.setObjectName('norelease_strip_tags')
        self.horizontalLayout.addWidget(self.norelease_strip_tags)
        self.vboxlayout.addLayout(self.horizontalLayout)
        self.verticalLayout.addWidget(self.groupBox)
        spacerItem = QtWidgets.QSpacerItem(368, 187, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)

        self.retranslateUi(NoReleaseOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(NoReleaseOptionsPage)

    def retranslateUi(self, NoReleaseOptionsPage):
        self.groupBox.setTitle(QtWidgets.QApplication.translate('NoReleaseOptionsPage', 'No release'))
        self.norelease_enable.setText(QtWidgets.QApplication.translate('NoReleaseOptionsPage', _('Enable plugin for all releases by default')))
        self.label.setText(QtWidgets.QApplication.translate('NoReleaseOptionsPage', _('Tags to strip (comma-separated)')))


def strip_release_specific_metadata(metadata):
    strip_tags = config.setting['norelease_strip_tags']
    strip_tags = [tag.strip() for tag in strip_tags.split(',')]
    for tag in strip_tags:
        metadata.delete(tag)


class NoReleaseAction(BaseAction):
    NAME = _('Remove specific release information...')

    def callback(self, objs):
        for album in objs:
            if isinstance(album, Album):
                strip_release_specific_metadata(album.metadata)
                for track in album.tracks:
                    strip_release_specific_metadata(track.metadata)
                    for file in track.linked_files:
                        track.update_file_metadata(file)
                album.update()


class NoReleaseOptionsPage(OptionsPage):
    NAME = 'norelease'
    TITLE = 'No release'
    PARENT = 'plugins'

    options = [
        BoolOption('setting', 'norelease_enable', False),
        TextOption('setting', 'norelease_strip_tags', 'asin,barcode,catalognumber,date,label,media,releasecountry,releasestatus'),
    ]

    def __init__(self, parent=None):
        super(NoReleaseOptionsPage, self).__init__(parent)
        self.ui = Ui_NoReleaseOptionsPage()
        self.ui.setupUi(self)

    def load(self):
        self.ui.norelease_strip_tags.setText(config.setting['norelease_strip_tags'])
        self.ui.norelease_enable.setChecked(config.setting['norelease_enable'])

    def save(self):
        config.setting['norelease_strip_tags'] = str(self.ui.norelease_strip_tags.text())
        config.setting['norelease_enable'] = self.ui.norelease_enable.isChecked()


def no_release_album_processor(tagger, metadata, release):
    if config.setting['norelease_enable']:
        strip_release_specific_metadata(metadata)


def no_release_track_processor(tagger, metadata, track, release):
    if config.setting['norelease_enable']:
        strip_release_specific_metadata(metadata)


register_album_metadata_processor(no_release_album_processor)
register_track_metadata_processor(no_release_track_processor)
register_album_action(NoReleaseAction())
register_options_page(NoReleaseOptionsPage)



PLUGIN_NAME = 'Standardise Feat.'
PLUGIN_AUTHOR = 'Sambhav Kothari'
PLUGIN_DESCRIPTION = 'Standardises "featuring" join phrases for artists to "feat."'
PLUGIN_VERSION = "0.3"
PLUGIN_API_VERSIONS = ["1.4", "2.0", "2.1", "2.2", "2.3"]
PLUGIN_LICENSE = "GPL-3.0"
PLUGIN_LICENSE_URL = "http://www.gnu.org/licenses/gpl-3.0.txt"

import re
from picard import log
from picard.metadata import register_track_metadata_processor, register_album_metadata_processor
from picard.plugin import PluginPriority

_feat_re = re.compile(r" f(ea)?t(\.|uring)? ", re.IGNORECASE)


def standardise_feat(artists_str, artists_list):
    """
    >>> standardise_feat("The A", ["The A"])
    'The A'
    >>> standardise_feat("The A ft. B", ["The A", "B"])
    'The A feat. B'
    >>> standardise_feat("The A & B featuring C", ["The A", "B", "C"])
    'The A & B feat. C'
    >>> standardise_feat("The A featuring B (and C)", ["The A", "B", "C"])
    'The A feat. B (and C)'
    """
    match_exp = r"(\s*.*\s*)".join((map(re.escape, artists_list))) + r"(\s*.*$)"
    try:
        join_phrases = re.match(match_exp, artists_str).groups()
    except AttributeError:
        log.debug("Unable to standardise artists: %r", artists_str)
        return artists_str
    else:
        standardised_join_phrases = [_feat_re.sub(" feat. ", phrase)
                                     for phrase in join_phrases]
        # Add a blank string at the end to allow zipping of
        # join phrases and artists_list since there is one less join phrase
        standardised_join_phrases.append("")
        return "".join([artist + join_phrase for artist, join_phrase in
                        zip(artists_list, standardised_join_phrases)])


def standardise_track_artist(tagger, metadata, track, release):
    metadata["artist"] = standardise_feat(metadata["artist"], metadata.getall("artists"))
    metadata["artistsort"] = standardise_feat(metadata["artistsort"], metadata.getall("~artists_sort"))


def standardise_album_artist(tagger, metadata, release):
    metadata["albumartist"] = standardise_feat(metadata["albumartist"], metadata.getall("~albumartists"))
    metadata["albumartistsort"] = standardise_feat(metadata["albumartistsort"], metadata.getall("~albumartists_sort"))


register_track_metadata_processor(standardise_track_artist, priority=PluginPriority.HIGH)
register_album_metadata_processor(standardise_album_artist, priority=PluginPriority.HIGH)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# This source code is a plugin for MusicBrainz Picard.
# It adds a context menu action to save files and rewrite their header.
# Copyright (C) 2015 Nicolas Cenerario
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.

from __future__ import unicode_literals

PLUGIN_NAME = "Save and rewrite header"
PLUGIN_AUTHOR = "Nicolas Cenerario"
PLUGIN_DESCRIPTION = "This plugin adds a context menu action to save files and rewrite their header."
PLUGIN_VERSION = "0.3"
PLUGIN_API_VERSIONS = ["0.9.0", "0.10", "0.15", "2.0"]
PLUGIN_LICENSE = "GPL-3.0-or-later"
PLUGIN_LICENSE_URL = "http://www.gnu.org/licenses/gpl-3.0.txt"

from _functools import partial
from mutagen import File as MFile
from picard.ui.itemviews import (BaseAction, register_album_action,
                                 register_cluster_action,
                                 register_clusterlist_action,
                                 register_track_action, register_file_action)
from picard.metadata import Metadata
from picard.util import thread


class save_and_rewrite_header(BaseAction):

    NAME = "Save and rewrite header"

    def __init__(self):
        super(save_and_rewrite_header, self).__init__()
        register_file_action(self)
        register_track_action(self)
        register_album_action(self)
        register_cluster_action(self)
        register_clusterlist_action(self)

    def callback(self, obj):
        def save(pf):
            metadata = Metadata()
            metadata.copy(pf.metadata)
            mf = MFile(pf.filename)
            if mf is not None:
                mf.delete()
            return pf._save_and_rename(pf.filename, metadata)
        for f in self.tagger.get_files_from_objects(obj, save=True):
            f.set_pending()
            thread.run_task(partial(save, f), f._saving_finished,
                            priority=2, thread_pool=f.tagger.save_thread_pool)

save_and_rewrite_header()



from picard import log
from picard.collection import Collection, user_collections
from picard.file import File, register_file_post_save_processor

from picard.plugins.add_to_collection import settings
from picard.plugins.add_to_collection.override_module import override_module


def post_save_processor(file: File) -> None:
    collection_id = settings.collection_id()
    if not collection_id:
        log.error("cannot find collection ID setting")
        return
    collection: Collection = user_collections.get(collection_id)
    if not collection:
        log.error(f"cannot find collection with id {collection_id}")
        return
    release_id = file.metadata["musicbrainz_albumid"]
    if release_id and release_id not in collection.releases:
        log.debug("Adding release %r to %r", release_id, collection.name)
        collection.add_releases(set([release_id]), callback=lambda: None)


def register_processor() -> None:
    with override_module(post_save_processor):
        register_file_post_save_processor(post_save_processor)



import enum
import json
from typing import Any, List, Mapping, Optional, Union

# This module is a perfect target for dataclasses.
# Picard though supports Python 3.5+, so we cannot
# use dataclasses for compatibility reasons.


class APIObject:
    """
    Base class from Deezer API objects.
    """
    fields = []  # type: List[str]

    def __init__(self, **kwargs):
        if len(self.fields) == 0:
            raise NotImplementedError(type(self).__name__ + ' cannot have an empty field list')
        for field in self.fields:
            setattr(self, field, kwargs.get(field))

    def __eq__(self, other):
        for field in self.fields:
            if getattr(self, field) != getattr(other, field):
                return False
        return True


class Artist(APIObject):
    """
    The Artist API object.
    """
    fields = ['name']


class CoverSize(enum.Enum):
    """
    Cover size selector.
    """
    THUMBNAIL = 'small'
    SMALL = ''
    MEDIUM = 'medium'
    BIG = 'big'
    LARGE = 'xl'


class Album(APIObject):
    """
    The Album API object.
    """
    fields = ['title', 'cover']

    def cover_url(self, cover_size: CoverSize) -> str:
        """
        Get the album cover URL based on the size wanted.
        """
        return '{}?size={}'.format(self.cover, cover_size.value)


class Track(APIObject):
    """
    The Track API object.
    """
    fields = ['album', 'artist']


available_objects = {c.__name__.lower(): c for c in APIObject.__subclasses__()}


def parse_json(data: Union[str, Mapping[str, Any]]) -> APIObject:
    if isinstance(data, str):
        return json.loads(data, object_hook=_dict_to_object)

    def convert_inner(data: Mapping[str, Any]):
        """
        Traverse the dictionaries with post-order
        algorithm to convert dict objects in Object instances.
        """
        for k, v in data.items():
            if isinstance(v, dict):
                data[k] = convert_inner(v)
        return _dict_to_object(data)

    convert_inner(data)
    return _dict_to_object(data)


def _dict_to_object(data: Mapping[str, Any]) -> Optional[APIObject]:
    try:
        obj_type = data['type']
        obj_class = available_objects[obj_type]
    except KeyError:
        return None
    else:
        return obj_class(**data)



# -*- coding: utf-8 -*-

# This is the Sort Multivalue Tags plugin for MusicBrainz Picard.
# Copyright (C) 2013 Sophist
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

PLUGIN_NAME = "Sort Multi-Value Tags"
PLUGIN_AUTHOR = "Sophist"
PLUGIN_DESCRIPTION = '''
This plugin sorts multi-value tags e.g. Performers alphabetically.<br /><br />
Note: Some multi-value tags are excluded for the following reasons:
<ol>
<li>Sequence is important e.g. Artists</li>
<li>The sequence of one tag is linked to the sequence of another e.g. Label and Catalogue number.</li>
</ol>
'''
PLUGIN_VERSION = "1.1"
PLUGIN_API_VERSIONS = ["0.15", "2.0"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from picard.metadata import register_track_metadata_processor
from picard.plugin import PluginPriority
from picard import log

# Define tags where sort order is important
_sort_multivalue_tags_exclude = (
    'artists', '~artists_sort', 'musicbrainz_artistid',
    '~albumartists', '~albumartists_sort', 'musicbrainz_albumartistid',
    'work', 'musicbrainz_workid',
    'label', 'catalognumber',
    'country', 'date',
    'releasetype',
    'mood',
)
# Possible future enhancement:
# Sort linked tags e.g. work so that the sequence in related tags e.g. workid retains the relationship between
# e.g. work and workid.


def sort_multivalue_tags(tagger, metadata, track, release):
    for tag in list(metadata.keys()):
        if tag in _sort_multivalue_tags_exclude:
            continue
        data = metadata.getall(tag)
        if len(data) > 1:
            sorted_data = sorted(data)
            if data != sorted_data:
                metadata.set(tag, sorted_data)
                log.debug("%s: Tag sorted: %s = %s", PLUGIN_NAME, tag, sorted_data)

                
register_track_metadata_processor(sort_multivalue_tags, priority=PluginPriority.LOW)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2018 Wieland Hoffmann
# Copyright (C) 2019-2023 Philipp Wolfer
# Copyright (C) 2020-2021 Laurent Monin
# Copyright (C) 2021 Bob Swift
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import importlib
import logging
import os
import shutil
import sys
import unittest
from tempfile import mkdtemp, mkstemp
from types import ModuleType
from typing import Any, Callable, Optional
from unittest.mock import Mock, patch

import picard
from picard import config, log
from picard.i18n import setup_gettext
from picard.plugin import _PLUGIN_MODULE_PREFIX, _unregister_module_extensions
from picard.pluginmanager import PluginManager
from picard.releasegroup import ReleaseGroup
from PyQt5 import QtCore


class FakeThreadPool(QtCore.QObject):

    def start(self, runnable, priority):
        runnable.run()


class FakeTagger(QtCore.QObject):

    tagger_stats_changed = QtCore.pyqtSignal()

    def __init__(self):
        QtCore.QObject.__init__(self)
        QtCore.QObject.config = config
        QtCore.QObject.log = log
        self.tagger_stats_changed.connect(self.emit)
        self.exit_cleanup = []
        self.files = {}
        self.stopping = False
        self.thread_pool = FakeThreadPool()
        self.priority_thread_pool = FakeThreadPool()
        self.save_thread_pool = FakeThreadPool()
        self.mb_api = Mock()

    def register_cleanup(self, func: Callable[[], Any]) -> None:
        self.exit_cleanup.append(func)

    def run_cleanup(self) -> None:
        for f in self.exit_cleanup:
            f()

    def emit(self, *args) -> None:
        pass

    def get_release_group_by_id(self, rg_id: str) -> ReleaseGroup:
        return ReleaseGroup(rg_id)


class PluginTestCase(unittest.TestCase):
    def setUp(self) -> None:
        log.set_level(logging.DEBUG)
        setup_gettext(None, "C")
        self.tagger = FakeTagger()
        QtCore.QObject.tagger = self.tagger
        QtCore.QCoreApplication.instance = lambda: self.tagger
        self.addCleanup(self.tagger.run_cleanup)
        self.init_config()

        self.tmp_directory = self.mktmpdir()
        # return tmp_directory from pluginmanager.plugin_dirs
        self.patchers = [
            patch(
                "picard.pluginmanager.plugin_dirs", return_value=[self.tmp_directory]
            ).start(),
            patch(
                "picard.util.thread.to_main",
                side_effect=lambda func, *args, **kwargs: func(*args, **kwargs),
            ).start(),
        ]

    def tearDown(self) -> None:
        for patcher in self.patchers:
            patcher.stop()

    @staticmethod
    def init_config() -> None:
        fake_config = Mock()
        fake_config.setting = {}
        fake_config.persist = {}
        fake_config.profiles = {}
        # Make config object available for legacy use
        config.config = fake_config
        config.setting = fake_config.setting
        config.persist = fake_config.persist
        config.profiles = fake_config.profiles

    @staticmethod
    def set_config_values(
        setting: Optional[dict] = None,
        persist: Optional[dict] = None,
        profiles: Optional[dict] = None,
    ) -> None:
        if setting:
            for key, value in setting.items():
                config.config.setting[key] = value
        if persist:
            for key, value in persist.items():
                config.config.persist[key] = value
        if profiles:
            for key, value in profiles.items():
                config.config.profiles[key] = value

    def mktmpdir(self, ignore_errors: bool = False) -> None:
        tmpdir = mkdtemp(suffix=self.__class__.__name__)
        self.addCleanup(shutil.rmtree, tmpdir, ignore_errors=ignore_errors)
        return tmpdir

    def copy_file_tmp(self, filepath: str, ext: Optional[str] = None) -> str:
        fd, copy = mkstemp(suffix=ext)
        os.close(fd)
        self.addCleanup(self.remove_file_tmp, copy)
        shutil.copy(filepath, copy)
        return copy

    @staticmethod
    def remove_file_tmp(filepath: str) -> None:
        if os.path.isfile(filepath):
            os.unlink(filepath)

    def _test_plugin_install(self, name: str, module: str) -> ModuleType:
        self.unload_plugin(module)
        with self.assertRaises(ImportError):
            importlib.import_module(f"picard.plugins.{module}")

        pm = PluginManager(plugins_directory=self.tmp_directory)

        msg = f"install_plugin: {name} from {module}"
        pm.install_plugin(os.path.join("plugins", module))
        self.assertEqual(len(pm.plugins), 1, msg)
        self.assertEqual(pm.plugins[0].name, name, msg)

        self.set_config_values(setting={"enabled_plugins": [module]})

        # if module is properly loaded, this should work
        return importlib.import_module(f"picard.plugins.{module}")

    def unload_plugin(self, plugin_name: str) -> None:
        """for testing purposes"""
        _unregister_module_extensions(plugin_name)
        if hasattr(picard.plugins, plugin_name):
            delattr(picard.plugins, plugin_name)
        key = _PLUGIN_MODULE_PREFIX + plugin_name
        if key in sys.modules:
            del sys.modules[key]



# -*- coding: utf-8 -*-

# Copyright (C) 2016 Anderson Mesquita <andersonvom@gmail.com>
# Copyright (C) 2019 Alan Swanson <reiver@improbability.net>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from picard import metadata

PLUGIN_NAME = "Hyphen unicode"
PLUGIN_AUTHOR = "Alan Swanson <revier@improbability.net>"
PLUGIN_VERSION = "1.0.1"
PLUGIN_API_VERSIONS = ["0.9", "0.10", "0.11", "0.15", "2.0"]
PLUGIN_LICENSE = "GPL-3.0-or-later"
PLUGIN_LICENSE_URL = "https://gnu.org/licenses/gpl.html"
PLUGIN_DESCRIPTION = '''Replaces unicode character HYPHEN (U+2010) [0xE2 0x80
0x90] with typographically identical HYPHEN-MINUS (U+002D) [0x2D] for fonts
that do not support HYPHEN and to prevent visually duplicate filenames
differentiated only by their hyphens.

Unicode duplicated hyphen from ASCII as an unambiguous way to designate a
hyphen from a minus whilst still being typographically indentical. Since
text processing on music tags is rare so choice is purely pedantic esepcially
as keyboards only have HYPHEN-MINUS.

Replaces character on "album", "title", "artist", "artists", "artistsort",
"albumartist", "albumartists" and "albumartistsort" tags.'''

# Based on Non-ASCII Equivalents plugin.
# Musicbrainz form discussion on HYPHEN versus HYPHEN-MINUS at;
# https://community.metabrainz.org/t/correct-hyphen-unicode-hyphen-or-hyphen-minus/19610

CHAR_TABLE = {
    # HYPHEN to HYPHEN-MINUS
    "‐": "-",
}

FILTER_TAGS = [
    "title",
    "artist",
    "artists",
    "artistsort",
    "album",
    "albumsort",
    "albumartist",
    "albumartists",
    "albumartistsort",
]


def sanitize(char):
    if char in CHAR_TABLE:
        return CHAR_TABLE[char]
    return char


def ascii(word):
    return "".join(sanitize(char) for char in word)


def main(tagger, metadata, *args):
    for name, value in metadata.rawitems():
        if name in FILTER_TAGS:
            metadata[name] = [ascii(x) for x in value]


metadata.register_track_metadata_processor(main)
metadata.register_album_metadata_processor(main)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/moodbars/ui_options_moodbar.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MoodbarOptionsPage(object):
    def setupUi(self, MoodbarOptionsPage):
        MoodbarOptionsPage.setObjectName("MoodbarOptionsPage")
        MoodbarOptionsPage.resize(305, 317)
        self.vboxlayout = QtWidgets.QVBoxLayout(MoodbarOptionsPage)
        self.vboxlayout.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout.setSpacing(6)
        self.vboxlayout.setObjectName("vboxlayout")
        self.moodbar_group = QtWidgets.QGroupBox(MoodbarOptionsPage)
        self.moodbar_group.setObjectName("moodbar_group")
        self.vboxlayout1 = QtWidgets.QVBoxLayout(self.moodbar_group)
        self.vboxlayout1.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout1.setSpacing(2)
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.label = QtWidgets.QLabel(self.moodbar_group)
        self.label.setObjectName("label")
        self.vboxlayout1.addWidget(self.label)
        self.vorbis_command = QtWidgets.QLineEdit(self.moodbar_group)
        self.vorbis_command.setObjectName("vorbis_command")
        self.vboxlayout1.addWidget(self.vorbis_command)
        self.label_2 = QtWidgets.QLabel(self.moodbar_group)
        self.label_2.setObjectName("label_2")
        self.vboxlayout1.addWidget(self.label_2)
        self.mp3_command = QtWidgets.QLineEdit(self.moodbar_group)
        self.mp3_command.setObjectName("mp3_command")
        self.vboxlayout1.addWidget(self.mp3_command)
        self.label_3 = QtWidgets.QLabel(self.moodbar_group)
        self.label_3.setObjectName("label_3")
        self.vboxlayout1.addWidget(self.label_3)
        self.flac_command = QtWidgets.QLineEdit(self.moodbar_group)
        self.flac_command.setObjectName("flac_command")
        self.vboxlayout1.addWidget(self.flac_command)
        self.label_4 = QtWidgets.QLabel(self.moodbar_group)
        self.label_4.setObjectName("label_4")
        self.vboxlayout1.addWidget(self.label_4)
        self.wav_command = QtWidgets.QLineEdit(self.moodbar_group)
        self.wav_command.setObjectName("wav_command")
        self.vboxlayout1.addWidget(self.wav_command)
        self.vboxlayout.addWidget(self.moodbar_group)
        spacerItem = QtWidgets.QSpacerItem(263, 21, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout.addItem(spacerItem)

        self.retranslateUi(MoodbarOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(MoodbarOptionsPage)

    def retranslateUi(self, MoodbarOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.moodbar_group.setTitle(_translate("MoodbarOptionsPage", "Moodbar"))
        self.label.setText(_translate("MoodbarOptionsPage", "Path to Ogg Vorbis moodbar tool:"))
        self.label_2.setText(_translate("MoodbarOptionsPage", "Path to MP3 moodbar tool:"))
        self.label_3.setText(_translate("MoodbarOptionsPage", "Path to FLAC moodbar tool:"))
        self.label_4.setText(_translate("MoodbarOptionsPage", "Path to WAV moodbar tool:"))



PLUGIN_NAME = 'Classic Disc Numbers'
PLUGIN_AUTHOR = 'Lukas Lalinsky'
PLUGIN_DESCRIPTION = '''Moves disc numbers and subtitles from the separate tags to album titles.'''
PLUGIN_VERSION = "0.2"
PLUGIN_API_VERSIONS = ["0.15", "2.0"]

from picard.metadata import register_track_metadata_processor
import re


def add_discnumbers(tagger, metadata, track, release):
    if int(metadata["totaldiscs"] or "0") > 1:
        if "discsubtitle" in metadata:
            metadata["album"] = "%s (disc %s: %s)" % (
                metadata["album"], metadata["discnumber"],
                metadata["discsubtitle"])
        else:
            metadata["album"] = "%s (disc %s)" % (
                metadata["album"], metadata["discnumber"])

register_track_metadata_processor(add_discnumbers)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/bpm/ui_options_bpm.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_BPMOptionsPage(object):
    def setupUi(self, BPMOptionsPage):
        BPMOptionsPage.setObjectName("BPMOptionsPage")
        BPMOptionsPage.resize(611, 273)
        self.verticalLayout = QtWidgets.QVBoxLayout(BPMOptionsPage)
        self.verticalLayout.setObjectName("verticalLayout")
        self.bpm_options = QtWidgets.QGroupBox(BPMOptionsPage)
        self.bpm_options.setObjectName("bpm_options")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.bpm_options)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalWidget = QtWidgets.QWidget(self.bpm_options)
        self.verticalWidget.setMaximumSize(QtCore.QSize(500, 16777215))
        self.verticalWidget.setObjectName("verticalWidget")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.verticalWidget)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.slider_parameter = QtWidgets.QSlider(self.verticalWidget)
        self.slider_parameter.setMinimum(1)
        self.slider_parameter.setMaximum(3)
        self.slider_parameter.setPageStep(1)
        self.slider_parameter.setOrientation(QtCore.Qt.Horizontal)
        self.slider_parameter.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.slider_parameter.setTickInterval(1)
        self.slider_parameter.setObjectName("slider_parameter")
        self.verticalLayout_3.addWidget(self.slider_parameter)
        self.slider_labels = QtWidgets.QHBoxLayout()
        self.slider_labels.setSpacing(6)
        self.slider_labels.setObjectName("slider_labels")
        self.slider_super_fast = QtWidgets.QLabel(self.verticalWidget)
        self.slider_super_fast.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.slider_super_fast.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.slider_super_fast.setObjectName("slider_super_fast")
        self.slider_labels.addWidget(self.slider_super_fast)
        self.slider_default = QtWidgets.QLabel(self.verticalWidget)
        self.slider_default.setToolTip("")
        self.slider_default.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTop|QtCore.Qt.AlignTrailing)
        self.slider_default.setObjectName("slider_default")
        self.slider_labels.addWidget(self.slider_default)
        self.verticalLayout_3.addLayout(self.slider_labels)
        self.line = QtWidgets.QFrame(self.verticalWidget)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout_3.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.samplerate_label = QtWidgets.QLabel(self.verticalWidget)
        self.samplerate_label.setObjectName("samplerate_label")
        self.gridLayout.addWidget(self.samplerate_label, 2, 0, 1, 1)
        self.samplerate_value = QtWidgets.QLabel(self.verticalWidget)
        self.samplerate_value.setText("")
        self.samplerate_value.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.samplerate_value.setObjectName("samplerate_value")
        self.gridLayout.addWidget(self.samplerate_value, 2, 1, 1, 1)
        self.hop_s_value = QtWidgets.QLabel(self.verticalWidget)
        self.hop_s_value.setText("")
        self.hop_s_value.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.hop_s_value.setObjectName("hop_s_value")
        self.gridLayout.addWidget(self.hop_s_value, 1, 1, 1, 1)
        self.hop_s_label = QtWidgets.QLabel(self.verticalWidget)
        self.hop_s_label.setObjectName("hop_s_label")
        self.gridLayout.addWidget(self.hop_s_label, 1, 0, 1, 1)
        self.win_s_value = QtWidgets.QLabel(self.verticalWidget)
        self.win_s_value.setText("")
        self.win_s_value.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.win_s_value.setObjectName("win_s_value")
        self.gridLayout.addWidget(self.win_s_value, 0, 1, 1, 1)
        self.win_s_label = QtWidgets.QLabel(self.verticalWidget)
        self.win_s_label.setObjectName("win_s_label")
        self.gridLayout.addWidget(self.win_s_label, 0, 0, 1, 1)
        self.gridLayout.setColumnStretch(0, 4)
        self.verticalLayout_3.addLayout(self.gridLayout)
        self.verticalLayout_2.addWidget(self.verticalWidget)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.verticalLayout.addWidget(self.bpm_options)

        self.retranslateUi(BPMOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(BPMOptionsPage)

    def retranslateUi(self, BPMOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.bpm_options.setTitle(_translate("BPMOptionsPage", "BPM Analyze Parameters:"))
        self.slider_super_fast.setText(_translate("BPMOptionsPage", "Super Fast"))
        self.slider_default.setText(_translate("BPMOptionsPage", "Default"))
        self.samplerate_label.setText(_translate("BPMOptionsPage", "Samplerate:"))
        self.hop_s_label.setText(_translate("BPMOptionsPage", "Number of frames between two consecutive runs:"))
        self.win_s_label.setText(_translate("BPMOptionsPage", "Length of FFT:"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/fanarttv/ui_options_fanarttv.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_FanartTvOptionsPage(object):
    def setupUi(self, FanartTvOptionsPage):
        FanartTvOptionsPage.setObjectName("FanartTvOptionsPage")
        FanartTvOptionsPage.resize(340, 317)
        self.vboxlayout = QtWidgets.QVBoxLayout(FanartTvOptionsPage)
        self.vboxlayout.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout.setSpacing(6)
        self.vboxlayout.setObjectName("vboxlayout")
        self.groupBox = QtWidgets.QGroupBox(FanartTvOptionsPage)
        self.groupBox.setObjectName("groupBox")
        self.vboxlayout1 = QtWidgets.QVBoxLayout(self.groupBox)
        self.vboxlayout1.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout1.setSpacing(2)
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.label = QtWidgets.QLabel(self.groupBox)
        self.label.setTextFormat(QtCore.Qt.RichText)
        self.label.setWordWrap(True)
        self.label.setOpenExternalLinks(True)
        self.label.setObjectName("label")
        self.vboxlayout1.addWidget(self.label)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout1.addItem(spacerItem)
        self.label_2 = QtWidgets.QLabel(self.groupBox)
        self.label_2.setObjectName("label_2")
        self.vboxlayout1.addWidget(self.label_2)
        self.fanarttv_client_key = QtWidgets.QLineEdit(self.groupBox)
        self.fanarttv_client_key.setObjectName("fanarttv_client_key")
        self.vboxlayout1.addWidget(self.fanarttv_client_key)
        self.vboxlayout.addWidget(self.groupBox)
        self.verticalGroupBox = QtWidgets.QGroupBox(FanartTvOptionsPage)
        self.verticalGroupBox.setObjectName("verticalGroupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalGroupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.fanarttv_cdart_use_always = QtWidgets.QRadioButton(self.verticalGroupBox)
        self.fanarttv_cdart_use_always.setObjectName("fanarttv_cdart_use_always")
        self.verticalLayout.addWidget(self.fanarttv_cdart_use_always)
        self.fanarttv_cdart_use_if_no_albumcover = QtWidgets.QRadioButton(self.verticalGroupBox)
        self.fanarttv_cdart_use_if_no_albumcover.setObjectName("fanarttv_cdart_use_if_no_albumcover")
        self.verticalLayout.addWidget(self.fanarttv_cdart_use_if_no_albumcover)
        self.fanarttv_cdart_use_never = QtWidgets.QRadioButton(self.verticalGroupBox)
        self.fanarttv_cdart_use_never.setObjectName("fanarttv_cdart_use_never")
        self.verticalLayout.addWidget(self.fanarttv_cdart_use_never)
        self.vboxlayout.addWidget(self.verticalGroupBox)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout.addItem(spacerItem1)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout.addItem(spacerItem2)

        self.retranslateUi(FanartTvOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(FanartTvOptionsPage)

    def retranslateUi(self, FanartTvOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.groupBox.setTitle(_translate("FanartTvOptionsPage", "fanart.tv cover art"))
        self.label.setText(_translate("FanartTvOptionsPage", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /></head><body>\n"
"<p>This plugin loads cover art from <a href=\"http://fanart.tv/\">fanart.tv</a>. If you want to improve the results of this plugin please contribute.</p>\n"
"<p>In order to use this plugin you have to register a personal API key on<br /><a href=\"https://fanart.tv/get-an-api-key/\">https://fanart.tv/get-an-api-key/</a></p></body></html>"))
        self.label_2.setText(_translate("FanartTvOptionsPage", "Enter your personal API key here:"))
        self.verticalGroupBox.setTitle(_translate("FanartTvOptionsPage", "Medium images"))
        self.fanarttv_cdart_use_always.setText(_translate("FanartTvOptionsPage", "Always load medium images"))
        self.fanarttv_cdart_use_if_no_albumcover.setText(_translate("FanartTvOptionsPage", "Load only if no front cover is available"))
        self.fanarttv_cdart_use_never.setText(_translate("FanartTvOptionsPage", "Never load medium images"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/wikidata/ui_options_wikidata.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_WikidataOptionsPage(object):
    def setupUi(self, WikidataOptionsPage):
        WikidataOptionsPage.setObjectName("WikidataOptionsPage")
        WikidataOptionsPage.resize(602, 512)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(WikidataOptionsPage)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.releaseGroup_groupBox = QtWidgets.QGroupBox(WikidataOptionsPage)
        self.releaseGroup_groupBox.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.releaseGroup_groupBox.sizePolicy().hasHeightForWidth())
        self.releaseGroup_groupBox.setSizePolicy(sizePolicy)
        self.releaseGroup_groupBox.setMinimumSize(QtCore.QSize(0, 0))
        self.releaseGroup_groupBox.setObjectName("releaseGroup_groupBox")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.releaseGroup_groupBox)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.use_release_group_genres = QtWidgets.QCheckBox(self.releaseGroup_groupBox)
        self.use_release_group_genres.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.use_release_group_genres.sizePolicy().hasHeightForWidth())
        self.use_release_group_genres.setSizePolicy(sizePolicy)
        self.use_release_group_genres.setChecked(True)
        self.use_release_group_genres.setObjectName("use_release_group_genres")
        self.verticalLayout_4.addWidget(self.use_release_group_genres)
        self.verticalLayout_2.addWidget(self.releaseGroup_groupBox)
        self.artist_groupBox = QtWidgets.QGroupBox(WikidataOptionsPage)
        self.artist_groupBox.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.artist_groupBox.sizePolicy().hasHeightForWidth())
        self.artist_groupBox.setSizePolicy(sizePolicy)
        self.artist_groupBox.setMinimumSize(QtCore.QSize(0, 0))
        self.artist_groupBox.setObjectName("artist_groupBox")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.artist_groupBox)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.use_artist_genres = QtWidgets.QCheckBox(self.artist_groupBox)
        self.use_artist_genres.setObjectName("use_artist_genres")
        self.verticalLayout_8.addWidget(self.use_artist_genres)
        self.hLayout_use_artist_no_release = QtWidgets.QHBoxLayout()
        self.hLayout_use_artist_no_release.setObjectName("hLayout_use_artist_no_release")
        spacerItem = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.hLayout_use_artist_no_release.addItem(spacerItem)
        self.use_artist_only_if_no_release = QtWidgets.QCheckBox(self.artist_groupBox)
        self.use_artist_only_if_no_release.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.use_artist_only_if_no_release.sizePolicy().hasHeightForWidth())
        self.use_artist_only_if_no_release.setSizePolicy(sizePolicy)
        self.use_artist_only_if_no_release.setCheckable(True)
        self.use_artist_only_if_no_release.setChecked(False)
        self.use_artist_only_if_no_release.setObjectName("use_artist_only_if_no_release")
        self.hLayout_use_artist_no_release.addWidget(self.use_artist_only_if_no_release)
        self.verticalLayout_8.addLayout(self.hLayout_use_artist_no_release)
        self.hLayout_ignore_genres_from_artists_label = QtWidgets.QHBoxLayout()
        self.hLayout_ignore_genres_from_artists_label.setObjectName("hLayout_ignore_genres_from_artists_label")
        spacerItem1 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.hLayout_ignore_genres_from_artists_label.addItem(spacerItem1)
        self.ignore_genres_from_these_artists_label = QtWidgets.QLabel(self.artist_groupBox)
        self.ignore_genres_from_these_artists_label.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ignore_genres_from_these_artists_label.sizePolicy().hasHeightForWidth())
        self.ignore_genres_from_these_artists_label.setSizePolicy(sizePolicy)
        self.ignore_genres_from_these_artists_label.setObjectName("ignore_genres_from_these_artists_label")
        self.hLayout_ignore_genres_from_artists_label.addWidget(self.ignore_genres_from_these_artists_label)
        self.verticalLayout_8.addLayout(self.hLayout_ignore_genres_from_artists_label)
        self.hLayout_ignore_genres_from_artists = QtWidgets.QHBoxLayout()
        self.hLayout_ignore_genres_from_artists.setObjectName("hLayout_ignore_genres_from_artists")
        spacerItem2 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.hLayout_ignore_genres_from_artists.addItem(spacerItem2)
        self.ignore_genres_from_these_artists = QtWidgets.QLineEdit(self.artist_groupBox)
        self.ignore_genres_from_these_artists.setEnabled(False)
        self.ignore_genres_from_these_artists.setObjectName("ignore_genres_from_these_artists")
        self.hLayout_ignore_genres_from_artists.addWidget(self.ignore_genres_from_these_artists)
        self.verticalLayout_8.addLayout(self.hLayout_ignore_genres_from_artists)
        self.verticalLayout_2.addWidget(self.artist_groupBox)
        self.work_groupBox = QtWidgets.QGroupBox(WikidataOptionsPage)
        self.work_groupBox.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.work_groupBox.sizePolicy().hasHeightForWidth())
        self.work_groupBox.setSizePolicy(sizePolicy)
        self.work_groupBox.setObjectName("work_groupBox")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.work_groupBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.use_work_genres = QtWidgets.QCheckBox(self.work_groupBox)
        self.use_work_genres.setChecked(True)
        self.use_work_genres.setObjectName("use_work_genres")
        self.verticalLayout_3.addWidget(self.use_work_genres)
        self.verticalLayout_2.addWidget(self.work_groupBox)
        self.generalSettings_groupBox = QtWidgets.QGroupBox(WikidataOptionsPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.generalSettings_groupBox.sizePolicy().hasHeightForWidth())
        self.generalSettings_groupBox.setSizePolicy(sizePolicy)
        self.generalSettings_groupBox.setMinimumSize(QtCore.QSize(0, 120))
        self.generalSettings_groupBox.setObjectName("generalSettings_groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.generalSettings_groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.ignore_genres_label = QtWidgets.QLabel(self.generalSettings_groupBox)
        self.ignore_genres_label.setObjectName("ignore_genres_label")
        self.verticalLayout.addWidget(self.ignore_genres_label)
        self.hLayout_ignore_these_genres = QtWidgets.QHBoxLayout()
        self.hLayout_ignore_these_genres.setObjectName("hLayout_ignore_these_genres")
        spacerItem3 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.hLayout_ignore_these_genres.addItem(spacerItem3)
        self.ignore_these_genres = QtWidgets.QLineEdit(self.generalSettings_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ignore_these_genres.sizePolicy().hasHeightForWidth())
        self.ignore_these_genres.setSizePolicy(sizePolicy)
        self.ignore_these_genres.setObjectName("ignore_these_genres")
        self.hLayout_ignore_these_genres.addWidget(self.ignore_these_genres)
        self.verticalLayout.addLayout(self.hLayout_ignore_these_genres)
        self.genre_delimiter_label = QtWidgets.QLabel(self.generalSettings_groupBox)
        self.genre_delimiter_label.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.genre_delimiter_label.sizePolicy().hasHeightForWidth())
        self.genre_delimiter_label.setSizePolicy(sizePolicy)
        self.genre_delimiter_label.setWordWrap(False)
        self.genre_delimiter_label.setObjectName("genre_delimiter_label")
        self.verticalLayout.addWidget(self.genre_delimiter_label)
        self.hLayout_genre_delimiter = QtWidgets.QHBoxLayout()
        self.hLayout_genre_delimiter.setObjectName("hLayout_genre_delimiter")
        spacerItem4 = QtWidgets.QSpacerItem(20, 17, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.hLayout_genre_delimiter.addItem(spacerItem4)
        self.genre_delimiter = QtWidgets.QComboBox(self.generalSettings_groupBox)
        self.genre_delimiter.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.genre_delimiter.sizePolicy().hasHeightForWidth())
        self.genre_delimiter.setSizePolicy(sizePolicy)
        self.genre_delimiter.setEditable(True)
        self.genre_delimiter.setObjectName("genre_delimiter")
        self.genre_delimiter.addItem("")
        self.genre_delimiter.setItemText(0, " / ")
        self.genre_delimiter.addItem("")
        self.genre_delimiter.addItem("")
        self.genre_delimiter.setItemText(2, ";")
        self.genre_delimiter.addItem("")
        self.genre_delimiter.addItem("")
        self.genre_delimiter.setItemText(4, ", ")
        self.hLayout_genre_delimiter.addWidget(self.genre_delimiter)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.hLayout_genre_delimiter.addItem(spacerItem5)
        self.verticalLayout.addLayout(self.hLayout_genre_delimiter)
        self.verticalLayout_2.addWidget(self.generalSettings_groupBox)
        spacerItem6 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem6)

        self.retranslateUi(WikidataOptionsPage)
        self.use_artist_genres.toggled['bool'].connect(self.ignore_genres_from_these_artists_label.setEnabled)
        self.use_artist_genres.toggled['bool'].connect(self.use_artist_only_if_no_release.setEnabled)
        self.use_artist_genres.toggled['bool'].connect(self.ignore_genres_from_these_artists.setEnabled)
        QtCore.QMetaObject.connectSlotsByName(WikidataOptionsPage)

    def retranslateUi(self, WikidataOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.releaseGroup_groupBox.setTitle(_translate("WikidataOptionsPage", "Release Group Genre Settings"))
        self.use_release_group_genres.setText(_translate("WikidataOptionsPage", "Use Release Group genres"))
        self.artist_groupBox.setTitle(_translate("WikidataOptionsPage", "Artist Genre Settings"))
        self.use_artist_genres.setText(_translate("WikidataOptionsPage", "Use Artist genres"))
        self.use_artist_only_if_no_release.setText(_translate("WikidataOptionsPage", "Use Artist genres only if no Release Group genres exist"))
        self.ignore_genres_from_these_artists_label.setText(_translate("WikidataOptionsPage", "Ignore Artist genres from these Artist(s):  (comma separated regular expressions)"))
        self.work_groupBox.setTitle(_translate("WikidataOptionsPage", "Work Genre Settings"))
        self.use_work_genres.setText(_translate("WikidataOptionsPage", "Use Work genres, when applicable"))
        self.generalSettings_groupBox.setTitle(_translate("WikidataOptionsPage", "General Settings"))
        self.ignore_genres_label.setText(_translate("WikidataOptionsPage", "Ignore these genres:   (comma separated regular expressions)"))
        self.genre_delimiter_label.setText(_translate("WikidataOptionsPage", "Genre Delimiter: (only applicable to ID3v2.3 tags)"))
        self.genre_delimiter.setItemText(1, _translate("WikidataOptionsPage", "; "))
        self.genre_delimiter.setItemText(3, _translate("WikidataOptionsPage", " ; "))



PLUGIN_NAME = "Add to Collection"
PLUGIN_AUTHOR = "Dvir Yitzchaki (dvirtz@gmail.com)"
PLUGIN_DESCRIPTION = "Adds any saved release to one of your user collections"
PLUGIN_VERSION = "0.1.1"
PLUGIN_API_VERSIONS = ["2.0"]
PLUGIN_LICENSE = ("MIT",)
PLUGIN_LICENSE_URL = "https://spdx.org/licenses/MIT.html"
PLUGIN_USER_GUIDE_URL = (
    "https://github.com/metabrainz/picard-plugins/blob/2.0/plugins/add_to_collection/README.md"
)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/post_tagging_actions/actions_status.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtWidgets


class Ui_ActionsStatus(object):
    def setupUi(self, ActionsStatus):
        ActionsStatus.setObjectName("ActionsStatus")
        ActionsStatus.resize(94, 18)
        self.horizontalLayout = QtWidgets.QHBoxLayout(ActionsStatus)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.actions_count = QtWidgets.QLabel(ActionsStatus)
        self.actions_count.setMinimumSize(QtCore.QSize(35, 0))
        self.actions_count.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.actions_count.setObjectName("actions_count")
        self.horizontalLayout.addWidget(self.actions_count)
        self.label = QtWidgets.QLabel(ActionsStatus)
        self.label.setText("")
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)

        self.retranslateUi(ActionsStatus)
        QtCore.QMetaObject.connectSlotsByName(ActionsStatus)

    def retranslateUi(self, ActionsStatus):
        _translate = QtCore.QCoreApplication.translate
        ActionsStatus.setWindowTitle(_translate("ActionsStatus", "Form"))
        self.actions_count.setText(_translate("ActionsStatus", "0"))
        self.label.setToolTip(_translate("ActionsStatus", "Remaining Actions"))



# -*- coding: utf-8 -*-

# This is the Decode Cyrillic plugin for MusicBrainz Picard.
# Copyright (C) 2015 aeontech
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function
PLUGIN_NAME = "Decode Cyrillic"
PLUGIN_AUTHOR = "aeontech"
PLUGIN_DESCRIPTION = '''
This plugin helps you quickly convert mis-encoded cyrillic Windows-1251 tags
to proper UTF-8 encoded strings. If your track/album names look something like
"Àëèñà â ñò›àíå ÷óäåñ", run this plugin from the context menu
before running the "Lookup" or "Scan" tools
'''
PLUGIN_VERSION = "1.1"
PLUGIN_API_VERSIONS = ["1.0", "2.0"]
PLUGIN_LICENSE = "MIT"
PLUGIN_LICENSE_URL = "https://opensource.org/licenses/MIT"

from picard import log
from picard.cluster import Cluster
from picard.ui.itemviews import BaseAction, register_cluster_action

_decode_tags = [
    'title',
    'albumartist',
    'artist',
    'album',
    'artistsort'
]
# _from_encoding = "latin1"
# _to_encoding   = "cp1251"


# TODO:
# - extend to support multiple codepage decoding, not just cp1251->latin1
#   instead, try the common variations, and show a dialog to the user,
#   allowing him to select the correct transcoding. See 2cyr.com for example.
# - also see http://stackoverflow.com/questions/23326531/how-to-decode-cp1252-string

class DecodeCyrillic(BaseAction):
    NAME = "Unmangle cyrillic metadata"

    def unmangle(self, tag, value):
        try:
            print(value, value.encode('latin1'))
            unmangled_value = value.encode('latin1').decode('cp1251')
        except UnicodeError:
            unmangled_value = value
            log.debug("%s: could not unmangle tag %s; original value: %s" % (PLUGIN_NAME, tag, value))
        return unmangled_value

    def callback(self, objs):
        for cluster in objs:
            if not isinstance(cluster, Cluster):
                continue

            for tag in _decode_tags:
                if not (tag in cluster.metadata):
                    continue

                cluster.metadata[tag] = self.unmangle(tag, cluster.metadata[tag])

            log.debug("cluster name is %s by %s" % (cluster.metadata['album'], cluster.metadata['albumartist']))

            for i, file in enumerate(cluster.files):

                log.debug("%s: Trying to unmangle file - original metadata %s" % (PLUGIN_NAME, file.orig_metadata))

                for tag in _decode_tags:

                    if not (tag in file.metadata):
                        continue

                    unmangled_tag = self.unmangle(tag, file.metadata[tag])

                    file.orig_metadata[tag] = unmangled_tag
                    file.metadata[tag] = unmangled_tag

                    file.orig_metadata.changed = True
                    file.metadata.changed = True
                    file.update(signal=True)

            cluster.update()

register_cluster_action(DecodeCyrillic())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './plugins/additional_artists_details/options_additional_artists_details.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AdditionalArtistsDetailsOptionsPage(object):
    def setupUi(self, AdditionalArtistsDetailsOptionsPage):
        AdditionalArtistsDetailsOptionsPage.setObjectName("AdditionalArtistsDetailsOptionsPage")
        AdditionalArtistsDetailsOptionsPage.resize(561, 802)
        AdditionalArtistsDetailsOptionsPage.setMinimumSize(QtCore.QSize(100, 0))
        self.verticalLayout = QtWidgets.QVBoxLayout(AdditionalArtistsDetailsOptionsPage)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea = QtWidgets.QScrollArea(AdditionalArtistsDetailsOptionsPage)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 543, 784))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gb_description = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.gb_description.setObjectName("gb_description")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.gb_description)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.format_description = QtWidgets.QLabel(self.gb_description)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.format_description.sizePolicy().hasHeightForWidth())
        self.format_description.setSizePolicy(sizePolicy)
        self.format_description.setWordWrap(True)
        self.format_description.setObjectName("format_description")
        self.verticalLayout_3.addWidget(self.format_description)
        self.verticalLayout_2.addWidget(self.gb_description)
        self.gb_process_track_artists = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.gb_process_track_artists.setObjectName("gb_process_track_artists")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.gb_process_track_artists)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label = QtWidgets.QLabel(self.gb_process_track_artists)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout_4.addWidget(self.label)
        self.cb_process_tracks = QtWidgets.QCheckBox(self.gb_process_track_artists)
        self.cb_process_tracks.setObjectName("cb_process_tracks")
        self.verticalLayout_4.addWidget(self.cb_process_tracks)
        self.verticalLayout_2.addWidget(self.gb_process_track_artists)
        self.gb_area_details = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.gb_area_details.setObjectName("gb_area_details")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.gb_area_details)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_2 = QtWidgets.QLabel(self.gb_area_details)
        self.label_2.setWordWrap(True)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_5.addWidget(self.label_2)
        self.cb_area_details = QtWidgets.QCheckBox(self.gb_area_details)
        self.cb_area_details.setObjectName("cb_area_details")
        self.verticalLayout_5.addWidget(self.cb_area_details)
        self.verticalLayout_2.addWidget(self.gb_area_details)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea)

        self.retranslateUi(AdditionalArtistsDetailsOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(AdditionalArtistsDetailsOptionsPage)

    def retranslateUi(self, AdditionalArtistsDetailsOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        AdditionalArtistsDetailsOptionsPage.setWindowTitle(_translate("AdditionalArtistsDetailsOptionsPage", "Form"))
        self.gb_description.setTitle(_translate("AdditionalArtistsDetailsOptionsPage", "Additional Artists Details"))
        self.format_description.setText(_translate("AdditionalArtistsDetailsOptionsPage", "<html><head/><body><p>These settings will determine how the <span style=\" font-weight:600;\">Additional Artists Details</span> plugin operates.</p><p>Please visit the repository on GitHub for <a href=\"https://github.com/rdswift/picard-plugins/blob/2.0_RDS_Plugins/plugins/additional_artists_details/docs/README.md\"><span style=\" text-decoration: underline; color:#0000ff;\">additional information</span></a>.</p></body></html>"))
        self.gb_process_track_artists.setTitle(_translate("AdditionalArtistsDetailsOptionsPage", "Process Track Artists"))
        self.label.setText(_translate("AdditionalArtistsDetailsOptionsPage", "<html><head/><body><p>This option determines whether or not details are retrieved for all track artists on the release. If you are only interested in details for the album artists then this should be disabled, thus significantly reducing the number of additional calls made to the MusicBrainz api and reducing the time required to load a release.  Album artists are always processed.</p></body></html>"))
        self.cb_process_tracks.setText(_translate("AdditionalArtistsDetailsOptionsPage", "Process track artists"))
        self.gb_area_details.setTitle(_translate("AdditionalArtistsDetailsOptionsPage", "Include Area Details"))
        self.label_2.setText(_translate("AdditionalArtistsDetailsOptionsPage", "<html><head/><body><p>This option determines whether or not County and Municipality information is included in the artist location variables created. Regardless of this setting, this information will be included if a County or Municipality is the area specified for an artist.</p></body></html>"))
        self.cb_area_details.setText(_translate("AdditionalArtistsDetailsOptionsPage", "Include area details"))



# -*- coding: utf-8 -*-

PLUGIN_NAME = "Generate Cuesheet"
PLUGIN_AUTHOR = "Lukáš Lalinský, Sambhav Kothari"
PLUGIN_DESCRIPTION = "Generate cuesheet (.cue file) from an album."
PLUGIN_VERSION = "1.2.2"
PLUGIN_API_VERSIONS = ["2.0"]


import os.path
import re
from PyQt5 import QtCore, QtWidgets
from picard.util import find_existing_path, encode_filename
from picard.ui.itemviews import BaseAction, register_album_action


_whitespace_re = re.compile(r'\s', re.UNICODE)
_split_re = re.compile(r'\s*("[^"]*"|[^ ]+)\s*', re.UNICODE)


def msfToMs(msf):
    msf = msf.split(":")
    return ((int(msf[0]) * 60 + int(msf[1])) * 75 + int(msf[2])) * 1000 / 75


class CuesheetTrack(list):

    def __init__(self, cuesheet, index):
        list.__init__(self)
        self.cuesheet = cuesheet
        self.index = index

    def set(self, *args):
        self.append(args)

    def find(self, prefix):
        return [i for i in self if tuple(i[:len(prefix)]) == tuple(prefix)]

    def getTrackNumber(self):
        return self.index

    def getLength(self):
        try:
            nextTrack = self.cuesheet.tracks[self.index + 1]
            index0 = self.find(("INDEX", "01"))
            index1 = nextTrack.find(("INDEX", "01"))
            return msfToMs(index1[0][2]) - msfToMs(index0[0][2])
        except IndexError:
            return 0

    def getField(self, prefix):
        try:
            return self.find(prefix)[0][len(prefix)]
        except IndexError:
            return ""

    def getArtist(self):
        return self.getField(("PERFORMER",))

    def getTitle(self):
        return self.getField(("TITLE",))

    def setArtist(self, artist):
        found = False
        for item in self:
            if item[0] == "PERFORMER":
                if not found:
                    item[1] = artist
                    found = True
                else:
                    del item
        if not found:
            self.append(("PERFORMER", artist))

    artist = property(getArtist, setArtist)


class Cuesheet(object):

    def __init__(self, filename):
        self.filename = filename
        self.tracks = []

    def read(self):
        with open(encode_filename(self.filename)) as f:
            self.parse(f.readlines())

    def unquote(self, string):
        if string.startswith('"'):
            if string.endswith('"'):
                return string[1:-1]
            else:
                return string[1:]
        return string

    def quote(self, string):
        if _whitespace_re.search(string):
            return '"' + string.replace('"', '\'') + '"'
        return string

    def parse(self, lines):
        track = CuesheetTrack(self, 0)
        self.tracks = [track]
        isUnicode = False
        for line in lines:
            # remove BOM
            if line.startswith('\xfe\xff'):
                isUnicode = True
                line = line[1:]
            # decode to unicode string
            line = line.strip()
            if isUnicode:
                line = line.decode('UTF-8', 'replace')
            else:
                line = line.decode('ISO-8859-1', 'replace')
            # parse the line
            split = [self.unquote(s) for s in _split_re.findall(line)]
            keyword = split[0].upper()
            if keyword == 'TRACK':
                trackNum = int(split[1])
                track = CuesheetTrack(self, trackNum)
                self.tracks.append(track)
            track.append(split)

    def write(self):
        lines = []
        for track in self.tracks:
            num = track.index
            for line in track:
                indent = 0
                if num > 0:
                    if line[0] == "TRACK":
                        indent = 2
                    elif line[0] != "FILE":
                        indent = 4
                line2 = " ".join([self.quote(s) for s in line])
                line2 = " " * indent + line2 + "\n"
                lines.append(line2.encode("UTF-8"))
        with open(encode_filename(self.filename), "wb") as f:
            f.writelines(lines)


class GenerateCuesheet(BaseAction):
    NAME = "Generate &Cuesheet..."

    def callback(self, objs):
        album = objs[0]
        current_directory = self.config.persist["current_directory"] or QtCore.QDir.homePath()
        current_directory = find_existing_path(str(current_directory))
        filename, selected_format = QtWidgets.QFileDialog.getSaveFileName(
            None, "", current_directory, "Cuesheet (*.cue)")
        if filename:
            cuesheet = Cuesheet(filename)
            #try: cuesheet.read()
            #except IOError: pass
            while len(cuesheet.tracks) <= len(album.tracks):
                track = CuesheetTrack(cuesheet, len(cuesheet.tracks))
                cuesheet.tracks.append(track)
            #if len(cuesheet.tracks) > len(album.tracks) - 1:
            #    cuesheet.tracks = cuesheet.tracks[0:len(album.tracks)+1]

            t = cuesheet.tracks[0]
            t.set("PERFORMER", album.metadata["albumartist"])
            t.set("TITLE", album.metadata["album"])
            t.set("REM", "MUSICBRAINZ_ALBUM_ID", album.metadata["musicbrainz_albumid"])
            t.set("REM", "MUSICBRAINZ_ALBUM_ARTIST_ID", album.metadata["musicbrainz_albumartistid"])
            if "date" in album.metadata:
                t.set("REM", "DATE", album.metadata["date"])
            index = 0.0
            for i, track in enumerate(album.tracks):
                mm = index / 60.0
                ss = (mm - int(mm)) * 60.0
                ff = (ss - int(ss)) * 75.0
                index += track.metadata.length / 1000.0
                t = cuesheet.tracks[i + 1]
                t.set("TRACK", "%02d" % (i + 1), "AUDIO")
                t.set("PERFORMER", track.metadata["artist"])
                t.set("TITLE", track.metadata["title"])
                t.set("REM", "MUSICBRAINZ_TRACK_ID", track.metadata["musicbrainz_trackid"])
                t.set("REM", "MUSICBRAINZ_ARTIST_ID", track.metadata["musicbrainz_artistid"])
                t.set("INDEX", "01", "%02d:%02d:%02d" % (mm, ss, ff))
                for file in track.linked_files:
                    audio_filename = file.filename
                    extension = audio_filename.split(".")[-1].lower()
                    if os.path.dirname(filename) == os.path.dirname(audio_filename):
                        audio_filename = os.path.basename(audio_filename)
                    if extension in ["mp3", "mp2", "m2a"]:
                        file_type = "MP3"
                    elif extension in ["aiff", "aif", "aifc"]:
                        file_type = "AIFF"
                    else:
                        file_type = "WAVE"
                    cuesheet.tracks[i].set("FILE", audio_filename, file_type)

            cuesheet.write()


action = GenerateCuesheet()
register_album_action(action)



PLUGIN_NAME = "Keep tags"
PLUGIN_AUTHOR = "Wieland Hoffmann"
PLUGIN_DESCRIPTION = """
Adds a $keep() function to delete all tags except the ones that you want.
Tags beginning with `musicbrainz_` are kept automatically, as are tags
beginning with `_`.

To keep all tags that can have a description (like `comment`, lyrics` and
`performer`), add `&lt;tagname without description&gt;` (not including `:`) to the
list of tags to keep."""

PLUGIN_VERSION = "1.2.1"
PLUGIN_API_VERSIONS = ["0.15.0", "0.15.1", "0.16.0", "1.0.0", "1.1.0", "1.2.0",
                       "1.3.0", "2.0"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from picard.script import register_script_function


@register_script_function
def keep(parser, *keeptags):
    tags = list(parser.context.keys())
    for tag in tags:
        if (tag in keeptags or
            tag.startswith("musicbrainz_") or
            tag.startswith("~")):
            continue
        if ":" in tag:
            tag_without_description = tag.split(":")[0]
            if tag_without_description in keeptags:
                continue
        parser.context.pop(tag, None)
    return ""
